package com.bitmechanic.maxq;

import java.sql.*;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Properties;

/**
 * Utility class used to simplify database operations from within
 * MaxQ python scripts.
 */
public class DBUtil
{
  private Connection conn;

  /**
   * Opens a database connection using the parameters provided.  Connection
   * is stored in an instance variable.  Call close() to close this
   * connection.
   *
   * @param driver   Class name of the JDBC driver to use
   * @param url      JDBC URL to connect to
   * @param username Database username to connect as
   * @param pw       Database password
   */
  public DBUtil(String driver, String url, String username, String pw)
      throws Exception
  {

    Class.forName(driver);
    conn = DriverManager.getConnection(url, username, pw);
  }

  /**
   * Oopens a database connection using the properties provided.  Connection
   * is stored in an instance variable.  Call close() to close this
   * connection.
   *
   * @param driver Class name of the JDBC driver to use
   * @param url    JDBC URL to connect to
   * @param props  JDBC properties to connect with
   */
  public DBUtil(String driver, String url, Properties props)
      throws Exception
  {

    Class.forName(driver);
    conn = DriverManager.getConnection(url, props);
  }

  /**
   * Returns the JDBC Connection object associated with this object.
   */
  public Connection getConnection()
  {
    return conn;
  }

  /**
   * Executes the INSERT/UPDATE/DELETE statement passed in.  Returns the
   * number of rows affected by the query.
   */
  public int execute(String sql) throws Exception
  {
    Statement stmt = conn.createStatement();
    int rows = stmt.executeUpdate(sql);
    stmt.close();

    return rows;
  }

  /**
   * Executes the SELECT statement passed in.  Returns an ArrayList of
   * HashMap objects.  One HashMap per row returned.  Keys in the HashMap
   * correspond to the columns in the SELECT statement.  If the ResultSet
   * is empty, a zero length ArrayList is returned.
   *
   * @param sql SELECT query to run
   */
  public ArrayList loadRows(String sql) throws Exception
  {
    Statement stmt = conn.createStatement();
    ResultSet rs = stmt.executeQuery(sql);
    ArrayList list = new ArrayList();
    ResultSetMetaData meta = rs.getMetaData();
    int cols = meta.getColumnCount();
    while (rs.next()) {
      HashMap map = rsToMap(rs, meta, cols);
      list.add(map);
    }
    rs.close();
    stmt.close();

    return list;
  }

  /**
   * Executes the SELECT statement passed in.  Returns a single HashMap
   * object corresponding to the first row in the ResultSet.  If the
   * ResultSet is empty, it returns null.
   *
   * @param sql SELECT query to run.
   */
  public HashMap loadRow(String sql) throws Exception
  {
    Statement stmt = conn.createStatement();
    ResultSet rs = stmt.executeQuery(sql);
    HashMap map = null;

    ResultSetMetaData meta = rs.getMetaData();
    int cols = meta.getColumnCount();
    if (rs.next()) {
      map = rsToMap(rs, meta, cols);
    }
    rs.close();
    stmt.close();

    return map;
  }

  /**
   * Executes the SELECT statement passed in.  Returns the first column of
   * the first row in the ResultSet.  If the ResultSet is empty, it returns
   * null.
   *
   * @param sql SELECT query to run
   */
  public String loadVal(String sql) throws Exception
  {
    Statement stmt = conn.createStatement();
    ResultSet rs = stmt.executeQuery(sql);
    String str = null;

    if (rs.next()) {
      str = rs.getString(1);
    }
    rs.close();
    stmt.close();

    return str;
  }


  /**
   * Closes the JDBC Connection associated with this instance.  Call this
   * when you're done with this object.  After calling this method, no
   * other methods will work.
   */
  public void close() throws Exception
  {
    conn.close();
  }

  private HashMap rsToMap(ResultSet rs, ResultSetMetaData meta, int cols)
      throws Exception
  {

    HashMap map = new HashMap();
    for (int i = 0; i < cols; i++) {
      int c = i + 1;
      map.put(meta.getColumnName(c), rs.getString(c));
    }

    return map;
  }
}
