package com.bitmechanic.maxq;

import com.bitmechanic.maxq.generator.GeneratorFactory;

import javax.swing.*;
import javax.swing.border.Border;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.io.IOException;

/**
 * The GUI for MaxQ.  The real work is delegated to currentTest, which is
 * what you might call our "document" (and we are the "view").
 */
public class SwingMain implements ActionListener
{
  private JMenu newFile = new JMenu("New");
  private JMenuItem openFile = new JMenuItem("Open...");
  private JMenuItem saveFile = new JMenuItem("Save");
  private JMenuItem saveFileAs = new JMenuItem("Save As...");
  //private JMenuItem prefs = new JMenuItem("Preferences...");
  private JMenuItem exit = new JMenuItem("Exit");
  private JMenuItem recordTest = new JMenuItem("Start Recording");
  private JMenuItem stopRecordTest = new JMenuItem("Stop Recording");
  private JMenuItem runTest = new JMenuItem("Run");

  private JFrame frame;
  private JMenuBar menuBar;
  private JLabel statusLabel;
  private JScrollPane editorPane;
  private JTextArea textArea;
  private Color origTextAreaBackground;

  private Test currentTest;
  private ProxyServer proxyServer;
  private static boolean outsideExceptionReportInProgress = false;
  private static SwingMain singleton;

  /**
   * One of these for each item in the File | New menu.  When fired, it
   * creates a new script with the appropriate generator.
   */
  private class NewGenerator extends JMenuItem implements ActionListener
  {
    String generatorClass;

    public NewGenerator(String className) throws Exception
    {
      super(GeneratorFactory.getClassDescription(className));
      generatorClass = className;
      addActionListener(this);
    }

    public void actionPerformed(ActionEvent e)
    {
      newFile(generatorClass);
    }
  }

  /**
   * Called once, to create the window.
   */
  public SwingMain(ProxyServer proxy) throws Exception
  {
    proxyServer = proxy;
    frame = new JFrame("MaxQ");
    editorPane = new JScrollPane();
    textArea = new JTextArea();
    textArea.setFont(new Font("Courier", Font.PLAIN, 12));
    textArea.setLineWrap(false);
    origTextAreaBackground = textArea.getBackground();
    textArea.setBackground(Color.LIGHT_GRAY);  // Until a file is opened.
    editorPane.setViewportView(textArea);

    Dimension minimumSize = new Dimension(600, 450);
    editorPane.setPreferredSize(minimumSize);

    JPanel statusPane = new JPanel();
    statusLabel = new JLabel(" - ", JLabel.LEFT);
    statusPane.setLayout(new GridLayout(1, 1));
    statusPane.add(statusLabel);
    Border border = BorderFactory.createCompoundBorder(BorderFactory.createRaisedBevelBorder(), BorderFactory.createLoweredBevelBorder());
    statusPane.setBorder(border);

    frame.getContentPane().add(editorPane, BorderLayout.CENTER);
    frame.getContentPane().add(statusPane, BorderLayout.SOUTH);

    // File menu
    JMenu fileMenu = new JMenu("File");
    String[] generators = GeneratorFactory.getClasses();
    int i;
    for (i = 0; i < generators.length; i++) {
      newFile.add(new NewGenerator(generators[i]));
    }
    fileMenu.add(newFile);
    fileMenu.add(openFile);
    fileMenu.add(saveFile);
    fileMenu.add(saveFileAs);
    fileMenu.addSeparator();
    fileMenu.add(exit);

    // Tools menu
    JMenu testMenu = new JMenu("Test");
    testMenu.add(recordTest);
    testMenu.add(stopRecordTest);
    testMenu.add(runTest);

    openFile.addActionListener(this);
    saveFile.addActionListener(this);
    saveFileAs.addActionListener(this);
    exit.addActionListener(this);
    recordTest.addActionListener(this);
    stopRecordTest.addActionListener(this);
    runTest.addActionListener(this);

    // Set menu bar
    menuBar = new JMenuBar();
    menuBar.add(fileMenu);
    menuBar.add(testMenu);
    frame.setJMenuBar(menuBar);
    updateGui();

    // There can be only one main window.  We make ourself available as a
    // static so that threads can report errors to us without knowing us.
    assert singleton == null;
    singleton = this;

    // Finish setting up the frame, and show it.
    frame.addWindowListener(new WindowAdapter()
    {
      public void windowClosing(WindowEvent e)
      {
        System.exit(0);
      }
    });
    frame.pack();
    frame.setVisible(true);
  }

  public void actionPerformed(ActionEvent e)
  {
    Object source = e.getSource();
    try {
      if (source == openFile) {
        openFile();
      } else if (source == saveFile) {
        saveFile();
      } else if (source == saveFileAs) {
        saveFileAs();
      } else if (source == runTest) {
        runTest();
      } else if (source == recordTest) {
        recordTest();
      } else if (source == stopRecordTest) {
        stopRecordTest();
      } else if (source == exit) {
        System.exit(0);
      }
    } catch (Exception ex) {
      reportException(ex);
    }
  }

  /**
   * Enables/disables menu items according to the state of the program.
   * Also sets the status bar according to the current situation.
   */
  private void updateGui()
  {
    // Just what is our state?
    boolean testOpen = currentTest != null;
    boolean recording = testOpen ? currentTest.isRecording() : false;

    // Enable/disable menu items.
    newFile.setEnabled(!recording);
    openFile.setEnabled(!recording);
    saveFile.setEnabled(testOpen);
    saveFileAs.setEnabled(testOpen);
    recordTest.setEnabled(testOpen && !recording);
    stopRecordTest.setEnabled(testOpen && recording);
    runTest.setEnabled(testOpen && !recording);

    // Enable the editor if there is a script loaded.
    textArea.setEnabled(testOpen);
    if (testOpen)
      textArea.setBackground(origTextAreaBackground);

    // Set status line.
    String status = " ";
    if (recording)
      status = "Recording Test";
    statusLabel.setText(status);
  }

  /**
   * Creates a new test using the supplied class for the generator.
   */
  private void newFile(String generatorClass)
  {
    setCurrentTest(new Test(proxyServer,
            new JTextAreaTestScriptAdapter(textArea),
            generatorClass));
    updateGui();
  }

  /**
   * Gets the starting directory for file open and save dialogs.
   */
  private String getCurrentPath()
  {
    if (currentTest == null)
      return System.getProperty("user.dir");
    else
      return currentTest.getCurrentPath();
  }

  /**
   * Shows the file open dialog and opens any script chosen.
   */
  private void openFile() throws Exception
  {
    final JFileChooser fc = new JFileChooser(getCurrentPath());
    int returnVal = fc.showOpenDialog(frame);

    if (returnVal == JFileChooser.APPROVE_OPTION) {
      openFile(fc.getSelectedFile());
    }
  }

  /**
   * Opens the script indicated by path for editing.
   */
  public void openFile(File path) throws Exception
  {
    setCurrentTest(new Test(proxyServer,
            new JTextAreaTestScriptAdapter(textArea),
            path));
    //frame.setTitle(currentTest.getTestName());
    updateGui();
  }

  private void setCurrentTest(Test test)
  {
    if (currentTest != null)
      currentTest.close();
    currentTest = test;
  }

  /**
   * Saves the script being edited to disk.
   */
  private void saveFile() throws IOException
  {
    if (currentTest.getTestFile() == null)
      saveFileAs();
    else
      currentTest.save();
  }

  /**
   * Shows the file save dialog and saves the script.
   */
  private void saveFileAs() throws IOException
  {
    final JFileChooser fc = new JFileChooser(currentTest.getCurrentPath());
    int returnVal = fc.showSaveDialog(frame);

    if (returnVal == JFileChooser.APPROVE_OPTION) {
      currentTest.setTestFile(fc.getSelectedFile());
      saveFile();
    }
  }

  /**
   * Starts recording.
   */
  private void recordTest()
  {
    currentTest.startRecording();
    updateGui();
  }

  /**
   * Stops recording.  Saves the test as a side effect.  This is desirable
   * because MaxQ is too stupid to prompt users to save on exit.
   */
  private void stopRecordTest() throws IOException
  {
    currentTest.stopRecording();
    saveFile();
    updateGui();
  }

  private void runTest()
  {
    statusLabel.setText("Running Test");
    new RunTestDialog(frame, textArea.getText());
    updateGui();
  }

  /**
   * Show an error report in a message box.
   */
  public void reportException(Throwable ex)
  {
    if (ex.getClass() != Utils.UserException.class)
      ex.printStackTrace();
    JOptionPane.showMessageDialog(frame, ex.getMessage(), "MaxQ Error",
        JOptionPane.ERROR_MESSAGE);
  }

  /**
   * Called by other threads to complain about their problems.  This better
   * than just dumping a stack trace because the user might not notice that.
   */
  public static void reportOutsideException(Throwable ex)
  {
    if (!outsideExceptionReportInProgress) {
      outsideExceptionReportInProgress = true;
      try {
        singleton.reportException(ex);
      } finally {
        outsideExceptionReportInProgress = false;
      }
    }
  }
}
