package com.bitmechanic.maxq;

import com.bitmechanic.maxq.generator.GeneratorFactory;
import com.bitmechanic.maxq.generator.IScriptGenerator;
import java.io.*;

/**
 * This encapsulates a test opened in the GUI.  It holds a script generator
 * and gets sent messages from the proxy server and the GUI and passes them
 * to it when recording is in progress.
 */
public class Test
{
  private IScriptGenerator generator;
  private IScriptAdapter scriptAdapter;
  private boolean isRecording;
  private File testFile;
  private String testName;
  ProxyServer proxyServer;

  /**
   * Create a new test using the supplied generator.
   */
  public Test(ProxyServer proxy, IScriptAdapter adapter,
              String generatorClass)
  {
    proxyServer = proxy;
    scriptAdapter = adapter;
    adapter.setText("");
    generator = GeneratorFactory.newGenerator(generatorClass, adapter);
    generator.doNew();
  }

  /**
   * Load an existing test from a file.
   */
  public Test(ProxyServer proxy, IScriptAdapter adapter, File file)
      throws IOException, Utils.UserException
  {
    proxyServer = proxy;
    scriptAdapter = adapter;
    testFile = file;

    // Load the script from the file.
    StringBuffer buffer = new StringBuffer();
    BufferedReader r = new BufferedReader(new FileReader(testFile));
    String line = r.readLine();
    while (line != null) {
      buffer.append(line);
      buffer.append(IScriptGenerator.EOL);
      line = r.readLine();
    }
    r.close();
    String script = buffer.toString();

    // Scripts must contain the class names of the generators that were
    // used to create them.
    String generatorClass = null;
    String[] classes = GeneratorFactory.getClasses();
    int i;
    for (i = 0; i < classes.length; i++) {
      if (script.indexOf(classes[i]) != -1) {
        generatorClass = classes[i];
        break;
      }
    }
    if (generatorClass == null)
      throw new Utils.UserException("Your file does not appear to have been generated by MaxQ.\n\n(MaxQ expects to find the class name of the generator somewhere in the script.\nPerhaps you have mistakenly deleted it?)");

    // Load the text of the script into the editor and attach the
    // appropriate generator.
    adapter.setText(script);
    generator = GeneratorFactory.newGenerator(generatorClass, adapter);
    generator.doLoad();
    testName = generator.parseTestName();
  }

  public String getTestName()
  {
    return testName;
  }

  public File getTestFile()
  {
    return testFile;
  }

  public void setTestFile(File file)
  {
    this.testFile = file;
  }

  public String getCurrentPath()
  {
    if (testFile == null)
      return System.getProperty("user.dir");
    else
      return stripFileName(testFile.getAbsolutePath());
  }

  /**
   * Returns a list of the valid file extensions for the test based on the code generator used.  For a jython
   * generator ".py" would be returned.
   *
   * @return String array containing valid extensions
   * @see IScriptGenerator#getValidFileExtensions
   */
  public String[] getvalidTestFileExtensions()
  {
    return generator.getValidFileExtensions();
  }

  public boolean isRecording()
  {
    return isRecording;
  }

  public void stopRecording()
  {
    proxyServer.removeObserver(generator);
    assert isRecording;
    generator.doStopRecording();
    isRecording = false;
  }

  public void startRecording()
  {
    assert !isRecording;
    proxyServer.addObserver(generator);
    isRecording = true;
    generator.doStartRecording();
  }

  public void save() throws IOException
  {
    generator.doSave(stripFileName(testFile.getAbsolutePath()), testFile.getName());

    PrintWriter writer = new PrintWriter(new FileWriter(testFile));
    String str = scriptAdapter.getText();
    str = Utils.replace(str, "\r\n", "\n");
    str = Utils.replace(str, "\n", IScriptGenerator.EOL);
    writer.println(str);
    writer.close();
  }

  private String stripFileName(String absolutePath)
  {
    int index = absolutePath.lastIndexOf(System.getProperty("file.separator"));
    if (index == -1) {
      return absolutePath;
    }
    return absolutePath.substring(0, index);
  }

  /**
   * Called when this instance is no longer required.
   */
  public void close()
  {
    if (isRecording)
      proxyServer.removeObserver(generator);
    generator.close();
  }
}
