package com.bitmechanic.maxq.generator;

import com.bitmechanic.maxq.*;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Abstract class providing general useful utility methods for code generation.
 *
 * @author Robert Dawson robert@rojotek.com
 * @author Wayne Hearn wayne_hearn@dell.com
 */
public abstract class AbstractCodeGenerator implements IScriptGenerator, ProxyServer.Observer, Runnable
{
  private Config config = Config.getConfig();
  // ------------------------------ FIELDS ------------------------------

  private static final Log log = LogFactory.getLog(AbstractCodeGenerator.class);

  private static String cpRspTo = System.getProperty("http.cpRspTo");
  private static boolean cpRspToStdout;
  private static boolean cpRspToFile = false;
  private static final String MIME_DEFAULTS[] = {"text/plain", "text/html", "text/comma-separated-values"};
  private static HashMap mimeTypes;
  protected static final String END_STATEMENT = ";" + EOL;


  /**
   * jtidyConfigFile variable serves two purpose:
   * one as a flag indicating we need spit out jtidy integration code,
   * and as a variable indicating the location of jtidy.cfg file.
   */
  protected static String jtidyConfigFile = System.getProperty("jtidy.config");

  protected long assertNumber = 0;
  protected boolean headersExist = false;

  private IScriptAdapter scriptAdapter;
  private String testName;
  private String defaultTestName;
  private String testPath;
  private boolean ignoreNextResponse;
  private Pattern[] namePatterns;
  private LinkedList outstandingInserts;
  private Thread insertThread;
  private String charset;

  private final Pattern insertMarkerRE = Pattern.compile("(^[ \\t]*" + EOL + ")?" + ".*\\^\\^\\^ Insert new recording.*", Pattern.MULTILINE);
  
  /**
   * timeElapsedSinceLastestRequest is the delay between the current and the
   * latest request processed.
   */
  private long timeElapsedSinceLastestRequest = 0;
  /**
   * Date of the latest request.
   */
  private Date latestRequestDate = new Date();
  /**
   * Is set to true if it is the first request of the test.
   */
  private boolean isFirstRequest = true;

  // -------------------------- STATIC METHODS --------------------------

  static
  {
    if (cpRspTo != null) {
      if (cpRspTo.toLowerCase().compareTo("stdout") == 0)
        cpRspToStdout = true;
      else if (cpRspTo.toLowerCase().compareTo("file") == 0)
        cpRspToFile = true;
    }

    mimeTypes = new HashMap();
    for (int i = 0; i < MIME_DEFAULTS.length; i++) {
      mimeTypes.put(MIME_DEFAULTS[i], MIME_DEFAULTS[i]);
    }
  }

  public static boolean isCpRspToStdout()
  {
    return cpRspToStdout;
  }

  public static boolean isCpRspToFile()
  {
    return cpRspToFile;
  }

  // --------------------------- CONSTRUCTORS ---------------------------

  /**
   * @param nameRegExps A list of regular expressions.  Each finds an
   *                    instance of the script's name embedded in the
   *                    script.  This class will use the regular expressions
   *                    to keep these names up-to-date when the user saves.
   */
  public AbstractCodeGenerator(IScriptAdapter adapter, String[] nameRegExps)
  {
    this.scriptAdapter = adapter;
    int i;
    namePatterns = new Pattern[nameRegExps.length];
    for (i = 0; i < nameRegExps.length; i++)
      namePatterns[i] = Pattern.compile(nameRegExps[i], Pattern.MULTILINE);

    testName = config.getProperty("test.default_testname", "MaxQTest");
    defaultTestName = testName;

    // Create a thread to allow multiple inserts to be collated and
    // inserted together.
    insertThread = new Thread(this);
    insertThread.start();
  }

  // --------------------- GETTER / SETTER METHODS ---------------------

  public String getDefaultTestName()
  {
    return defaultTestName;
  }

  public void setDefaultTestName(String defaultTestName)
  {
    this.defaultTestName = defaultTestName;
  }

  public IScriptAdapter getScriptAdapter()
  {
    return scriptAdapter;
  }

  /**
   * Returns the name of the test.
   * Currently the same as the name of the file the test is stored in
   *
   * @return testName
   */
  public String getTestName()
  {
    return testName;
  }

  /**
   * Sets the name of the test class.
   * Defaults to value set in maxq.properties under test.default_testname
   *
   * @param testName
   */
  public void setTestName(String testName)
  {
    this.testName = testName;
  }

  /**
   * Returns the file path of the test.  Null if test is not previously saved.
   */
  public String getTestPath()
  {
    return testPath;
  }

  /**
   * Sets the file path of the test
   *
   * @param testPath
   */
  public void setTestPath(String testPath)
  {
    this.testPath = testPath;
  }

  /**
   * Returns time elapsed since the latest request.
   */
  public long getTimeElapsedSinceLastestRequest() {
    return timeElapsedSinceLastestRequest;
  }
  
  public boolean isFirstRequest() {
    return isFirstRequest;
  }

  // ------------------------ INTERFACE METHODS ------------------------


  // --------------------- Interface IScriptGenerator ---------------------


  public String parseTestName()
  {
    Matcher m = namePatterns[0].matcher(getScript());
    if (!m.find()) {
      throw new IllegalArgumentException("You have fiddled with the formatting of the name of the test embedded in this file.  I can no longer understand it.");
    }
    return m.group(1);
  }

  /**
   * When a script is saved we update the name of the script embedded
   * within it.  Usually this is the class name for the script.
   */
  public void doSave(String path, String fileName)
  {
    // We do not include any extension in out test name.
    String name = fileName;
    int dotPos = fileName.indexOf(".");
    if (dotPos > -1)
      name = fileName.substring(0, dotPos);
    setTestName(name);
    setTestPath(path);

    // Update the name of the test in the file because it may have changed.
    int i;
    for (i = 0; i < namePatterns.length; i++) {
      Matcher m = namePatterns[i].matcher(getScript());
      if (!m.find())
        throw new IllegalArgumentException("Regular expression (\"" + namePatterns[i].pattern() + "\") provided to constructor cannot match the contents of the script");
      scriptAdapter.replace(name, m.start(1), m.end(1));
    }
  }

  public void doLoad()
  {
    setTestName(parseTestName());
  }

  /**
   * Necessary because the thread started in the constructor will otherwise
   * never exit.
   */
  public void close()
  {
    insertThread.interrupt();
  }

  // --------------------- Interface Observer ---------------------


  /**
   * @param header
   * @param requestBody
   */
  public void processRequest(HttpRequestHeader header, byte[] requestBody) throws Exception
  {
    // set the delay
    Date d = new Date();
    timeElapsedSinceLastestRequest = d.getTime() - latestRequestDate.getTime();
    latestRequestDate = d; // date update for the next time...
  
    if (header.method.toLowerCase().equals("get") && urlIgnored(header.url)) {
      log.debug("Ignoring GET request: \"" + header.url + "\"");
      ignoreNextResponse = true;
      return;
    }

    boolean isMultiPartRequest = header.contentType.startsWith("multipart/form-data");
    Action action = new Action();
    
    int pos = header.contentType.indexOf(';');
    if (pos != -1) {
      String newCharset = header.contentType.substring(pos);
      if (!newCharset.equals(charset))
        doSetCharset(newCharset);
    } else {
      if (charset != null)
        doSetCharset(null);
    }

    doHeaders(createHeaders(header));

    // parse out query string
    String url = header.url;
    log.debug("  recording url: " + url);

    String method = header.method.toLowerCase();

    pos = url.indexOf("?");
    if (pos != -1 && (url.length() > pos + 1)) {
      setQueryStringParams(action, url.substring(pos + 1));
      action.setUrl(HttpTestCase.replaceURL(url.substring(0, pos)));
    } else
      action.setUrl(HttpTestCase.replaceURL(url));

    if (requestBody.length > 0) {
      if (isMultiPartRequest) {
        // Update the action with the mutipart data.
        setMultiPartData(action, header, requestBody);
      } else {
        setBodyParams(action, new String(requestBody));
      }
    }

    // For compatibility with old generators which don't implement 
    // doQueryStringParameterList and doBodyParameterList, I call 
    // doParameterList with all parameters (query string and body
    // parameters merged).
    if (action.getParamsCount() > 0) {
      Param params[] = action.getParams();
      // Decode URL-encoded parameters if necessary.
      // I would have expected this to have been done at an earlier
      // stage, but I'm too lazy to fix it.  OCB.
      for (int i = 0; i < params.length; i++) {
        String name = params[i].name;
        String value = params[i].value;
        if (name.indexOf("+") != -1 || name.indexOf("%") != -1)
          params[i].name = HttpTestCase.staticUrlDecode(name);
        if (value.indexOf("+") != -1 || value.indexOf("%") != -1)
          params[i].value = HttpTestCase.staticUrlDecode(value);
      } 	
      doParameterList(params);
    }
    
    // Query string parameters.
    if (action.getQueryStringParamCount() > 0) {
      Param params[] = action.getQueryStringParams();
      // Decode URL-encoded parameters if necessary.
      // I would have expected this to have been done at an earlier
      // stage, but I'm too lazy to fix it.  OCB.
      for (int i = 0; i < params.length; i++) {
        String name = params[i].name;
        String value = params[i].value;
        if (name.indexOf("+") != -1 || name.indexOf("%") != -1)
          params[i].name = HttpTestCase.staticUrlDecode(name);
        if (value.indexOf("+") != -1 || value.indexOf("%") != -1)
          params[i].value = HttpTestCase.staticUrlDecode(value);
      }
      doQueryStringParameterList(params);
    }
    
    // Body parameters.
    if (action.getBodyParamCount() > 0) {
      Param params[] = action.getBodyParams();
      // Decode URL-encoded parameters if necessary.
      // I would have expected this to have been done at an earlier
      // stage, but I'm too lazy to fix it.  OCB.
      for (int i = 0; i < params.length; i++) {
        String name = params[i].name;
        String value = params[i].value;
        if (name.indexOf("+") != -1 || name.indexOf("%") != -1)
          params[i].name = HttpTestCase.staticUrlDecode(name);
        if (value.indexOf("+") != -1 || value.indexOf("%") != -1)
          params[i].value = HttpTestCase.staticUrlDecode(value);
      }
      doBodyParameterList(params);
    }

    // inform the user which URL is being tested
    String url_str = header.url;
    if (requestBody.length > 0)
      url_str += '?' + new String(requestBody);
    //TODO: The following could be part of MultiPart requests if
    //formatted appropriately.
    if (!isMultiPartRequest) {
      doTestUrlMessage(HttpTestCase.staticUrlDecode(url_str).trim());
    }

    //These fields default to empty for non Multipart requests.
    String data_str = "";
    String cont_len_str = "";

    //Compile the multipart-specific data.
    if (isMultiPartRequest) {
      method = method + "MultiPart";
      data_str = ", data";
      cont_len_str = ", " + header.contentLength;

      doSetData(action.getMultiPartData());
    }

    doCallUrl(action.getUrl(), method, data_str, cont_len_str);
  }

  private HeaderValue[] createHeaders(HttpRequestHeader header)
  {
    Vector v = new Vector();
    if (header.contentType.length() > 0)
      v.add(new HeaderValue(HeaderValue.ENTITY_CONTENT_TYPE, header.contentType));
    if (header.accept.length() > 0)
      v.add(new HeaderValue(HeaderValue.REQUEST_ACCEPT, header.accept));
    if (header.referer.length() > 0)
      v.add(new HeaderValue(HeaderValue.REQUEST_REFERER, header.referer));
    if (header.pragmaNoCache)
      v.add(new HeaderValue(HeaderValue.GENERAL_CACHE_CONTROL, "no-cache"));
    HeaderValue[] hv = new HeaderValue[v.size()];
    return (HeaderValue[]) v.toArray(hv);
  }

  /**
   * Processes the response and generates test events for the code generator implementation class.
   *
   * @param header
   * @param response
   */
  public void processResponse(HttpRequestHeader header, byte[] response) throws Exception
  {
    if (ignoreNextResponse) {
      log.debug("Ignoring response");
      ignoreNextResponse = false;
      return;
    }

    String respStr = new String(response).toLowerCase();

    String respCode = parseResponseCode(respStr);
    String contentType = parseContentType(respStr);

    doAssertResponse(respCode);

    if (contentType != null && mimeTypes.get(contentType) != null && respCode != null) {
      if (respCode.startsWith("200")) {
        if (respStr.indexOf("org.apache.struts.taglib.html.token") > 0){
          setStruts(true);
        } else {
          setStruts(false);
        }
        if (contentType.toLowerCase().compareTo("text/html") == 0) {
          doTidyCode(HttpTestCase.staticUrlDecode(header.url));
          if (cpRspToStdout) {
            doResponseForStdOut(HttpTestCase.staticUrlDecode(header.url).trim());
          } else if (cpRspToFile) {
            doResponseForFile();
          }
        } //text/html
      }// respCode.startsWith("200")
      doEndTransaction();      // add in an empty line for readability
    } else {
      log.debug("Ignoring response because content type is not known: " + contentType);
    }

    if (isFirstRequest())
      isFirstRequest = false;
  }

  // --------------------- Interface Runnable ---------------------


  /**
   * A thread for performing insertions.  Insertions can be slow when the file
   * gets large, but this allows them to be grouped together.
   */
  public void run()
  {
    // Initialization.
    outstandingInserts = new LinkedList();
    try {
      while (!Thread.interrupted()) {
        try {
          StringBuffer sb;
          synchronized (outstandingInserts)
          {
            outstandingInserts.wait();
            sb = new StringBuffer();
            while (outstandingInserts.size() > 0)
              sb.append((String) outstandingInserts.removeFirst());

            // Figure out where we will insert this.
            Matcher m = insertMarkerRE.matcher(getScriptAdapter().getText());
            if (!m.find())
              throw new Utils.UserException("You have altered your script and it no longer includes the lines that you were told not to alter.");
            getScriptAdapter().insert(sb.toString(), m.start(0));
          }
        } catch (Utils.UserException e) {
          SwingMain.reportOutsideException(e);
        }
      }
    } catch (InterruptedException e) {
      log.trace("Thread dead");
    }
  }

  // -------------------------- OTHER METHODS --------------------------

  /*
   * Routines for our descendents.
   */

  public abstract void doAssertResponse(String respCode) throws Utils.UserException;

  public abstract void doCallUrl(String url, String method, String data, String contentLength) throws Utils.UserException;

  /**
   * Called after all other calls for a particular transaction (request +
   * response).
   */
  public abstract void doEndTransaction() throws Utils.UserException;
  
  /**
   * Called if the processResponse find that the response
   * contains the Struts token
   */
  public abstract void setStruts(boolean struts);

  /**
   * Called when the character set for POST requests changes.
   */
  public abstract void doSetCharset(String cs) throws Utils.UserException;

  /**
   * Implementer must add script code for the parameters provided.
   * This will only be called if there is at least one parameter.
   *
   * @param params Query string and body parameters, without distinction.
   */
  public abstract void doParameterList(Param[] params) throws Utils.UserException;
  
  /**
   * @param params Query string parameters.
   */
  public abstract void doQueryStringParameterList(Param[] params) throws Utils.UserException;
  
  /** 
   * @param params Body parameters.
   */
  public abstract void doBodyParameterList(Param[] params) throws Utils.UserException;

  public abstract void doResponseForFile() throws Utils.UserException;

  public abstract void doResponseForStdOut(String url) throws Utils.UserException;

  public abstract void doSetData(String data) throws Utils.UserException;

  // Not yet documented.
  public abstract void doTestUrlMessage(String url) throws Utils.UserException;

  public abstract void doTidyCode(String url) throws Utils.UserException;

  public abstract void doHeaders(HeaderValue[] headers);

  /**
   * Returns the full script from the ScriptAdapter.
   */
  protected String getScript()
  {
    return this.scriptAdapter.getText();
  }

  /**
   * Inserts new lines into the script.  The lines are inserted before
   * a special comment that is included in the output of doNew().  This
   * means that recording can be stopped and restarted, even after the
   * file has been saved and reloaded.
   */
  protected void insert(String s) throws Utils.UserException
  {
    synchronized (outstandingInserts)
    {
      outstandingInserts.add(s);
      outstandingInserts.notifyAll();
    }
  }

  private String parseContentType(String response)
  {
    String contentType = null;

    int pos = response.indexOf("content-type:");  // Already lower case.
    if (pos > -1) {
      pos += 14;  // okay we found 'content-type:', advance 14 char positions
      int end = response.indexOf("\r\n", pos);
      int end2 = response.indexOf(";", pos);
      if (end2 > -1 && end2 < end) end = end2;
      if (end > -1) {
        contentType = response.substring(pos, end).trim();
      }
      log.debug("  Content-Type: " + contentType);
    } else {
      log.debug("  No content-type header!  First few lines:");
      StringTokenizer st = new StringTokenizer(response, "\n");
      int i = 0;
      while (st.hasMoreTokens() && i < 5) {
        log.debug(st.nextToken());
        i++;
      }
    }
    return contentType;
  }

  private String parseResponseCode(String response)
  {
    // parse the HTTP response code
    String respCode = null;
    int pos = response.indexOf(" ");
    if (pos != -1) {
      int end = response.indexOf(" ", pos + 1);

      // check for newline too
      int end2 = response.indexOf("\n", pos + 1);
      if (end2 != -1 && end2 < end) end = end2;
      if (end != -1) respCode = response.substring(pos + 1, end).trim();
    }

    log.debug("HTTP response code: " + respCode);
    return respCode;
  }

  /**
   * Updates the given Action with the multipart form-data that
   * is part of the current Request.
   * <p/>
   * Note that I have better things to do than read the HTTP RFC so I'm
   * only implementing what I can see is necessary for accepting form
   * data based on examples sent by my browser.  oliverbock.
   */
  private void setMultiPartData(Action action, HttpRequestHeader header, byte[] strarray)
  {
    // Extract the boundary between parts.
    String str = new String(strarray);
    int begin = header.contentType.indexOf("boundary=");
    int end = header.contentType.indexOf("; ", begin);
    if (end == -1)
      end = header.contentType.length();
    String boundary = header.contentType.substring(begin + 9, end);

    // Parse a parameter from each part.
    String[] parts = str.split("--" + boundary);
    assert parts[0].equals("");   // Separator begins data.
    int i;
    Pattern re = Pattern.compile("\r\nContent-Disposition: form-data; name=\"([^\"]+)\"[^\r\n]*\r\n\r\n(.*)\r\n", Pattern.DOTALL);
    for (i = 1; i < parts.length - 1; i++) {
      Matcher m = re.matcher(parts[i]);
      boolean ok = m.matches();
      assert ok;
      Param p = new Param();
      p.name = m.group(1);
      p.value = m.group(2);
      action.addBodyParam(p);
    }
    // Observed terminator.
    assert parts[parts.length - 1].equals("--\r\n");

    /*
         * I have no idea what the below code was intended to do, but since
         * this was previously never called (due to isMultiPartRequest always
         * being false) I felt it was safe to comment it out.  I just want
         * maxq to work with multipart/form-data.  oliverbock.
         *
         * // Trim off any unnecessary leading white spaces.
         * str = str.trim();
         * // Escape any single quotes found in the form-data.
         * str = Utils.replace(str, "'", "\\'");
         *
         * StringBuffer msb = new StringBuffer();
         * for (int i = 0; i < str.length(); i++) {
         *     if (str.charAt(i) == '\n') {
         *         msb.append("\\r\\n'\\\n        '");
         *     } else if (str.charAt(i) == '\r') {
         *         //skip the '\r' since we have to add the escape character
         *     } else {
         *         msb.append(str.charAt(i));
         *     }
         * }
         * action.setMultiPartData(msb.toString());
         */
  }

  private void setQueryStringParams(Action action, String str)
  {
    List params = Utils.getParams(str);
    Iterator iter = params.iterator();
    while (iter.hasNext()) {
      action.addQueryStringParam((Param) iter.next());
    }
  }

  private void setBodyParams(Action action, String str)
  {
    List params = Utils.getParams(str);
    Iterator iter = params.iterator();
    while (iter.hasNext()) {
      action.addBodyParam((Param) iter.next());
    }
  }

  private boolean urlIgnored(String str)
  {
    // If "include" patterns are given, then a match must occur to
    //  not-ignore the request.
    // If "exclude" patterns are given, then a match must NOT occur to
    //  not-ignore the request.
    // Even if a match occurs with "include" patterns, an "exclude" pattern
    //  must still NOT match to not-ignore the request.
    boolean ignore = false;
    //    Iterator iter;
    List incPats = config.getIncludePatterns();
    List excPats = config.getExcludePatterns();
    if (!ignore && incPats.size() != 0) {
      log.debug("urlIgnored: checking for include matches");
      ignore = checkMatch(incPats.iterator(), str, false, true);
    }
    if (!ignore && excPats.size() != 0) {
      log.debug("urlIgnored: checking for exclude matches");
      ignore = checkMatch(excPats.iterator(), str, true, ignore);
    }
    return ignore;
  }

  private boolean checkMatch(Iterator iter, String str, boolean res, boolean def)
  {
    while (iter.hasNext()) {
      Matcher m = ((Pattern) iter.next()).matcher(str);
      log.debug("checkMatch: for \"" + str + "\" with pattern: \"" + m.pattern().pattern() + "\"");
      if (m.find()) {
        log.debug("checkMatch: Found match, returning: " + res);
        return res;
      }
    }
    log.debug("checkMatch: Didn't find match, returning: " + def);
    return def;
  }
}