package com.bitmechanic.maxq.generator;

import com.bitmechanic.maxq.Config;
import com.bitmechanic.maxq.IScriptAdapter;
import com.bitmechanic.maxq.Utils;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

/**
 * Factory for IScriptGenerator instances.
 * This loads files using configuration information found in the
 * properties file specified by the system property
 * generator.properties.file.
 * <p/>
 * If no file is specified the configuration will look for a file called
 * maxq.properties, and if this isn't found, the
 * JythonCodeGenerator will be used.
 * <p/>
 * The properties file will be used to specify the class name of the IScriptGenerator to be used,
 * and any properties that are specific to the class.
 *
 * @author Robert Dawson robert@rojotek.com
 */
public class GeneratorFactory
{
  private static Config config = Config.getConfig();

  /**
   * property specifying the name of the test script class to use.
   * The test script class is any class that implements the
   * <code>IScriptGenerator</code> interface
   */
  public static final String GENERATOR_CLASSNAMES_PROPERTY = "generator.classnames";

  /**
   * generator that outputs Jython code.
   */
  public static final String JYTHON_CLASSNAME = "com.bitmechanic.maxq.generator.JythonCodeGenerator";

  private GeneratorFactory()
  {
  }

  /**
   * Returns a list of class names to offer the user as generators.
   */
  public static String[] getClasses()
  {
    String names = config.getProperty(GENERATOR_CLASSNAMES_PROPERTY,
        JYTHON_CLASSNAME);
    return Utils.splitString(names, ":");
  }

  /**
   * Returns a description of the generator in className, for use in the
   * GUI.
   */
  public static String getClassDescription(String className)
      throws ClassNotFoundException, IllegalAccessException,
      InvocationTargetException
  {
    Class gen = Class.forName(className);
    Method getGenDesc;
    try {
      getGenDesc = gen.getMethod("getGeneratorDescription", null);
    } catch (NoSuchMethodException e) {
      // It is acceptable that this method not be implemented.
      return className;
    }
    return (String) getGenDesc.invoke(null, null);
  }

  /**
   * Retrieves the test script implementation object.  Use the property generator.classname
   * from the properties file.
   *
   * @return Return a IScriptGenerator instance, configured using the properties file.
   */
  public static IScriptGenerator newGenerator(String className,
                                              IScriptAdapter adapter)
  {
    Class clazz = null; //everyone needs a variable called clazz ;)
    try {
      clazz = GeneratorFactory.class.getClassLoader().loadClass(className);
    } catch (ClassNotFoundException e) {
      throw new IllegalStateException("Invalid test script class specified .  Class not found:" + className);
    }

    //if clazz implements the test script interface then instantiate it
    if (IScriptGenerator.class.isAssignableFrom(clazz)) {
      try {
        Class[] generatorConstructorParameterClasses = {IScriptAdapter.class};
        Object[] generatorConstructorParameters = {adapter};
        Constructor constructor = clazz.getDeclaredConstructor(generatorConstructorParameterClasses);
        IScriptGenerator generator = (IScriptGenerator) constructor.newInstance(generatorConstructorParameters);
        return generator;
      } catch (NoSuchMethodException e) {
        throw new IllegalStateException("Invalid test script class specified.  Must have a constructor that takes in properties: " + className);
      } catch (InstantiationException e) {
        throw new IllegalStateException("Invalid test script class specified.  Could not construct: " + className);
      } catch (InvocationTargetException e) {
        e.printStackTrace();
        throw new IllegalStateException("Invalid test script class specified.  Exception thrown in constructor: " + className);
      } catch (IllegalAccessException e) {
        throw new IllegalStateException("Invalid test script class specified.  IllegalAccessException: " + className);
      }
    } else {
      throw new IllegalStateException("Invalid test script class specified.  Does not implement IScriptGenerator: " + className);
    }
  }
}