// 
// Copyright (c) 2005, Caltha - Gajda, Krzewski, Mach, Potempski Sp.J. 
// All rights reserved. 
// 
// Redistribution and use in source and binary forms, with or without modification,  
// are permitted provided that the following conditions are met: 
//  
// * Redistributions of source code must retain the above copyright notice,  
//	 this list of conditions and the following disclaimer. 
// * Redistributions in binary form must reproduce the above copyright notice,  
//	 this list of conditions and the following disclaimer in the documentation  
//	 and/or other materials provided with the distribution. 
// * Neither the name of the Caltha - Gajda, Krzewski, Mach, Potempski Sp.J.  
//	 nor the names of its contributors may be used to endorse or promote products  
//	 derived from this software without specific prior written permission. 
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"  
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED  
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
// IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,  
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  
// BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
// OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,  
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE  
// POSSIBILITY OF SUCH DAMAGE. 
// 
package com.bitmechanic.maxq.generator;

import java.io.IOException;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.xml.sax.SAXException;

import com.meterware.httpunit.Button;
import com.meterware.httpunit.FrameSelector;
import com.meterware.httpunit.HTMLElementPredicate;
import com.meterware.httpunit.HTMLSegment;
import com.meterware.httpunit.SubmitButton;
import com.meterware.httpunit.WebClient;
import com.meterware.httpunit.WebConversation;
import com.meterware.httpunit.WebForm;
import com.meterware.httpunit.WebLink;
import com.meterware.httpunit.WebResponse;

/**
 * {@link TemplateLogic} variation for generating smart HttpUnit tests.
 * 
 * @author <a href="mailto:rafal@caltha.pl">Rafal Krzewski</a>
 * @version $Id: HttpUnitTemplateLogic.java,v 1.3 2005/05/04 12:45:44 rkrzewsk Exp $
 */
public class HttpUnitTemplateLogic extends TemplateLogic {

    // -- page parsing -------------------------------------------------------------------------- 
    
    /**
     * Parse a web page using HttpUnit.
     * 
     * <p>At the time of writing this requires patched HttpUnit to operate correctly</p>
     * 
     * @param url the page URL, required for downloading scripts.
     * @param content the the page content.
     */
    public HTMLSegment parsePage(String url, String content) throws MalformedURLException {
        return new InPlaceResponse(webClient, new URL(url), content);
    }

    /**
     * WebClient instance to be used for downloading scripts. 
     */
    private final WebClient webClient = new WebConversation();

    /**
     * A subclass of WebResponse for parsing pages intercepted by MaxQ proxy.
     * 
     * Based on WebResponse.DefaultResponse from HttpUnit project.
     */
    public static class InPlaceResponse extends WebResponse {

        /**
         * Creates new InPlaceResponse instance.
         * 
         * @param client the WebClient to use for downloading additional content (scripts).
         * @param url the page base url.
         * @param text the page content.
         */
        InPlaceResponse(WebClient client, URL url, String text) {
            this(client, FrameSelector.TOP_FRAME, url, text);
        }

        /**
         * Creates new InPlaceResponse instance.
         * 
         * @param client the WebClient to use for downloading additional content (scripts).
         * @param frame the selected frame (see HttpUnit docs)
         * @param url the page base url.
         * @param text the page content.
         */
        InPlaceResponse(WebClient client, FrameSelector frame, URL url,
                String text) {
            super(client, frame, url, text);
        }

        /**
         * Returns the response code associated with this response.
         */
        public int getResponseCode() {
            return HttpURLConnection.HTTP_OK;
        }

        /**
         * Returns the response message associated with this response.
         */
        public String getResponseMessage() {
            return "OK";
        }

        /**
         * Returns the header names.
         */
        public String[] getHeaderFieldNames() {
            return new String[] { "Content-type" };
        }

        /**
         * Returns the value for the specified header field. If no such field is defined, will
         * return null.
         */
        public String getHeaderField(String fieldName) {
            if (fieldName.equalsIgnoreCase("Content-type")) {
                return "text/html; charset=us-ascii";
            } else {
                return null;
            }
        }

        /**
         * Returns the values of the specified header.
         */
        public String[] getHeaderFields(String fieldName) {
            String value = getHeaderField(fieldName);
            return value == null ? new String[0] : new String[] { value };
        }

        /**
         * Returns a string representation of the object.
         */
        public String toString() {
            try {
                return "InPlaceResponse [" + getText() + "]";
            } catch (IOException e) { // should never happen
                return "InPlaceResponse [???]";
            }
        }
    }

    // -- GET heuristics ------------------------------------------------------------------------
    
    /**
     * Matches links whose hrefs are contained the specified criteria string.
     */
    private static final HTMLElementPredicate MATCH_URL_STRING_INVERSE = new HTMLElementPredicate() {
        public boolean matchesCriteria(Object htmlElement, Object criteria) {
            return ((String) criteria).endsWith(((WebLink) htmlElement).getURLString());
        };
    };

    /**
     * Detect links that have hrefs contained in the specified URL.
     * 
     * @param page the page.
     * @param url the url.
     * @return link array.
     * @throws SAXException if page traversal fails.
     */    
    private WebLink[] getLinksByUrl(HTMLSegment page, String url)
        throws SAXException
    {
        WebLink[] links = page.getMatchingLinks(
            MATCH_URL_STRING_INVERSE, url);
        return links;
    }
    
    /**
     * Returns list of IDs of the links that have hrefs containd in the specified URL.
     * 
     * @param page the page.
     * @param url the url.
     * @return list of link IDs.
     * @throws SAXException if page traversal fails.
     */
    // List<String>
    public List getLinkIds(HTMLSegment page, String url) throws SAXException {
        WebLink[] links = getLinksByUrl(page, url);
        List result = new ArrayList(links.length); // List<String>
        for(int i = 0; i < links.length; i++) {
            if(links[i].getID() != null && links[i].getID().length() > 0)
                result.add(links[i].getID());
        }
        return result;
    }

    /**
     * Returns list of texts of the links that have hrefs containd in the specified URL.
     * 
     * @param page the page.
     * @param url the url.
     * @return list of link texts.
     * @throws SAXException if page traversal fails.
     */
    // List<String>
    public List getLinkTexts(HTMLSegment page, String url) throws SAXException {
        WebLink[] links = getLinksByUrl(page, url);
        List result = new ArrayList(links.length); // List<String>
        for(int i = 0; i < links.length; i++) {
            result.add(links[i].getText());
        }
        return result;
    }

    // -- POST heuristics - form detection  -----------------------------------------------------

    /**
     * Returns all the forms contained in the given page.
     * 
     * @param page the page.
     * @throws SAXException if page traversal fails.
     */
    // List<WebForm>
    public List getForms(HTMLSegment page) throws SAXException {
        // silly wrapper around static method
        return Arrays.asList(page.getForms());
    }        

    /**
     * Guesses the submitted form using the names of the POST parameters in the next request.
     * 
     * @param page the page
     * @param requestParameters the POST parameters in the next request.
     * @return a form object, or null if not matched.
     * @throws SAXException if page traversal fails.
     */
    public WebForm getFormByParameterNames(HTMLSegment page,
        Map requestParameters) throws SAXException {
        WebForm[] forms = page.getForms();
        for (int i = 0; i < forms.length; i++) {
            Collection formParameters = Arrays.asList(forms[i]
                    .getParameterNames());
            if (formParameters.containsAll(requestParameters.keySet())) {
                return forms[i];
            }
        }
        return null;
    }    
    
    // -- POST heuristics -----------------------------------------------------------------------

    /**
     * Guesses the clicked submit button in a form, based on the POST parameters in the next 
     * request.
     * 
     * @param form the form.
     * @param parameters the POST parameters in the next request.
     * @return a submit button object, or null if not matched.
     * @throws SAXException if page traversal fails.
     */
    public SubmitButton getPostSubmitButton(WebForm form, Map parameters) 
        throws SAXException {
        SubmitButton[] submits = form.getSubmitButtons();
        for(int j = 0; j < submits.length; j++) {
            if(submits[j].getValue().equals(parameters.get(submits[j].getName())))
                return submits[j];
        }
        return null;
    }

    /**
     * Returns a list of Integers from 0 to max-1.
     * 
     * @param max number of Integers in the list.
     * @return the list of Integers.
     */
    private List numbers(int max) {
        List result = new ArrayList(max);
        for(int i = 0; i < max; i++)
            result.add(new Integer(i));
        return result;
    }
    
    /**
     * Returns the indices of submit buttons present in the form.
     * 
     * <p>At the time of writiting this method requires patched HttpUnit to compile</p>
     * 
     * @param form the form.
     * @return indices of all the submit buttons present in the form.
     */
    // List<String>
    public List getPostSubmitButtonIndices(WebForm form) {
        SubmitButton[] submitButtons = form.getSubmitButtons();
        if(submitButtons[0].isFake()) 
            return null;
        else
            return numbers(submitButtons.length);
    }

    /**
     * Returns all buttons in the form that can possibly submit it using JavaScritpt.
     * 
     * @param form the form.
     * @return list of Button objects.
     */
    // List<Button>
    private List getJSButtons(WebForm form) {
        Button[] buttons = form.getButtons();
        List result = new ArrayList();
        for(int i = 0; i < buttons.length; i++) {
            if(buttons[i].getAttribute("onclick").length() > 0 || 
                            buttons[i].getAttribute("onchange").length() > 0)
                result.add(buttons[i]);
        }
        return result;
    }

    /**
     * Returns list of IDs of the buttons that can possibly submit a form using JavaScript.
     * 
     * @param form the form.
     * @return list of button IDs.
     */
    // List<String>
    public List getPostButtonId(WebForm form) {
        List jsButtons = getJSButtons(form);
        List result = new ArrayList();
        for(Iterator i = jsButtons.iterator(); i.hasNext();) {
            Button button = (Button)i.next();
            if(button.getID().length() > 0) 
                result.add(button.getID());
        }
        return null;
    }

    /**
     * Returns list of indices of the buttons that can possibly submit a form using JavaScript.
     * 
     * @param form the form.
     * @return list of button indices.
     */
    // List<String>
    public List getPostButtonIndices(WebForm form) {
        Button[] buttons = form.getButtons();
        List jsButtons = getJSButtons(form);
        List result = new ArrayList(jsButtons.size());
        for(int i = 0; i < buttons.length; i++) {
            if(jsButtons.contains(buttons[i]))
                result.add(new Integer(i));
        }
        return result;
    }
    
    /**
     * Returns all links in the page that execute JavaScript code.
     * 
     * @param page the page.
     * @return list of link objects.
     * @throws SAXException if page traversal fails.
     */
    // List<WebLink>
    private List getJSLinks(HTMLSegment page) throws SAXException {
        WebLink[] links = page.getLinks();
        List result = new ArrayList();
        for(int i = 0; i < links.length; i++) {
            if(links[i].getAttribute("href").startsWith("javascript:") || 
                            links[i].getAttribute("onclick").length() > 0)
                result.add(links[i]);
        }
        return result;
    }
    
    /**
     * Returns IDs of all links in the page that execute JavaScript code.
     * 
     * @param page the page.
     * @return list of link IDs.
     * @throws SAXException if page traversal fails.
     */
    // List<String>
    public List getPostLinkIds(HTMLSegment page) throws SAXException {
        List jsLinks = getJSLinks(page);
        List result = new ArrayList();
        for(Iterator i = jsLinks.iterator(); i.hasNext();) {
            WebLink link = (WebLink)i.next();
            if(link.getID().length() > 0) 
                result.add(link.getID());
        }
        return result;
    }
    
    /**
     * Returns texts of all links in the page that execute JavaScript code.
     * 
     * @param page the page.
     * @return list of link texts.
     * @throws SAXException if page traversal fails.
     */
    // List<String>
    public List getPostLinkTexts(HTMLSegment page) throws SAXException {
        List jsLinks = getJSLinks(page);
        List result = new ArrayList();
        for(Iterator i = jsLinks.iterator(); i.hasNext();) {
            WebLink link = (WebLink)i.next();
            result.add(link.getText());
        }
        return result;
    }    
}
