package com.bitmechanic.maxq.generator;

import com.bitmechanic.maxq.HeaderValue;
import com.bitmechanic.maxq.IScriptAdapter;
import com.bitmechanic.maxq.Param;
import com.bitmechanic.maxq.Utils;
import com.bitmechanic.maxq.Utils.UserException;
import com.bitmechanic.maxq.Config;


/**
 * Generates an ISAC scenario for CLIF Load Injection Framework
 * @link http://clif.ow2.org/
 * 
 * @author Olivier Lecrivain olivier.lecrivain@gmail.com
 * @author Francois Vinassac f.vinassac@gmail.com
 * @author Bruno Dillenseger
 */
public class IsacCodeGenerator extends AbstractCodeGenerator {

  static final String TIMER_PROP = "generator.isac.timer";
  static final String NULL_TIMER = "null"; 
  static final String CONSTANT_TIMER = "ConstantTimer";
  static final String RANDOM_TIMER = "Random";
  static final String DEFAULT_TIMER = CONSTANT_TIMER;
  static final String RANDOM_TIMER_DIST_PROP = "generator.isac.timer.random.dist";
  static final String RANDOM_TIMER_DIST_UNIFORM = "uniform";
  static final String RANDOM_TIMER_DIST_GAUSSIAN = "gaussian";
  static final String RANDOM_TIMER_DIST_POISSON = "poisson";
  static final String RANDOM_TIMER_DIST_NEGEXPO = "negexpo";
  static final String RANDOM_TIMER_DIST_DEFAULT = RANDOM_TIMER_DIST_UNIFORM;
  static final String RANDOM_TIMER_MIN_PROP = "generator.isac.timer.random.min";
  static final String RANDOM_TIMER_DELTA_PROP = "generator.isac.timer.random.delta";
  static final String RANDOM_TIMER_UNIT_PROP = "generator.isac.timer.random.unit";
  static final String RANDOM_TIMER_DEVIATION_PROP = "generator.isac.timer.random.deviation"; 
  static final String[][] escapes = { // caution: order cares!
    { "\\", "\\\\" }, // first
    { ";", "\\;" },   // second
    { "&", "&amp;" }, // third
    { "\"", "&quot;" },
    { "<", "&lt;" },
    { ">", "&gt;" }
  };

  private String queryStringParameters = "";

  private String bodyParameters = "";

  private String timer, randomDist;
  
  private int delta=0, deviation=0, unit=1;

  static private String escapeXmlString(String str)
  {
    for (int i=0 ; i<escapes.length ; ++i) {
      str = str.replace(escapes[i][0], escapes[i][1]);
    }
    return str;
  }

  static private String encodedParameterList(Param[] params)
  {
	String result = "";
    for (int i = 0; i < params.length; i++) {
      if (i > 0) {
    	result += ";";
      }
      result += escapeXmlString(params[i].name) + "=" + escapeXmlString(params[i].value);
    }
    return result;
  }

  public IsacCodeGenerator(IScriptAdapter adapter) {
    super(adapter, new String[] {});
  }

  /**
   * Return the description of the Generator. This description appears in the
   * GUI, as "ISAC scenario"
   * 
   * @return The description of the Generator
   */
  public static String getGeneratorDescription() {
    return "ISAC scenario for CLIF.ow2.org";
  }

  /*
   * (non-Javadoc)
   * 
   * @see com.bitmechanic.maxq.generator.IScriptGenerator#doNew()
   */
  public void doNew() {
    wl("<?xml version=\"1.0\"?>");
    wl("<!DOCTYPE scenario SYSTEM \"classpath:org/ow2/clif/scenario/isac/dtd/scenario.dtd\">");
    wl("<scenario>");
    wl("	<behaviors>");
    wl("		<plugins>");
    wl("			<use id=\"replayHttp\" name=\"HttpInjector\">");
    wl("				<params>");
    wl("					<param name=\"indepthload\" value=\"\"></param>");
    wl("					<param name=\"proxyuserpass\" value=\"\"></param>");
    wl("					<param name=\"useragent\" value=\"\"></param>");
    wl("					<param name=\"proxyport\" value=\"\"></param>");
    wl("					<param name=\"proxyhost\" value=\"\"></param>");
    wl("					<param name=\"proxyusername\" value=\"\"></param>");
    wl("					<param name=\"preemptiveauthentication\" value=\"\"></param>");
    wl("				</params>");
    wl("			</use>");

    Config conf = Config.getConfig();
    timer = conf.getProperty(TIMER_PROP, DEFAULT_TIMER);

    if (timer.equals(CONSTANT_TIMER)) {
      wl("			<use id=\"replayTimer\" name=\"ConstantTimer\"></use>");
    } else if (timer.equals(RANDOM_TIMER)) {
      wl("      <use id=\"replayTimer\" name=\"Random\"></use>");
      randomDist = Config.getConfig().getProperty(RANDOM_TIMER_DIST_PROP, RANDOM_TIMER_DIST_DEFAULT);
      try {
	    if (randomDist.equals(RANDOM_TIMER_DIST_UNIFORM)) {
	      delta = conf.getPropertyInt(RANDOM_TIMER_DELTA_PROP).intValue();
	    }
	    else if (randomDist.equals(RANDOM_TIMER_DIST_GAUSSIAN)) {
	      delta = conf.getPropertyInt(RANDOM_TIMER_DELTA_PROP).intValue();
	      deviation = conf.getPropertyInt(RANDOM_TIMER_DEVIATION_PROP).intValue();
	    }
	    else if (randomDist.equals(RANDOM_TIMER_DIST_POISSON)) {
	      unit = conf.getPropertyInt(RANDOM_TIMER_UNIT_PROP).intValue();
	    }
	    else if (randomDist.equals(RANDOM_TIMER_DIST_NEGEXPO)) {
	      delta = conf.getPropertyInt(RANDOM_TIMER_DELTA_PROP).intValue();
	    }
      } catch (UserException ex)
      {
    	ex.printStackTrace(System.err);
    	System.err.println("Warning: incorrect property settings for Isac random timer distribution " + randomDist);
      }
    }

    wl("		</plugins>");
    wl("		<behavior id=\"session\">");
    wl("");
    wl("		<!-- ^^^ Insert new recordings here.  (Do not remove this line.) -->");
    wl("");
    wl("		</behavior>");
    wl("	</behaviors>");
    wl("	<loadprofile></loadprofile>");
    wl("</scenario>");
  }

  public String[] getValidFileExtensions() {
    return new String[] { ".xis" };
  }

  public void doParameterList(Param[] params) throws UserException {
  }

  public void doQueryStringParameterList(Param[] params) throws UserException {
	queryStringParameters = encodedParameterList(params);
  }

  public void doBodyParameterList(Param[] params) throws UserException {
	bodyParameters = encodedParameterList(params);
  }

  public void doTestUrlMessage(String url) {
  }

  public void doSetData(String data) {
//    wl("doSetData(data = " + data + ")");
  }

  public void doCallUrl(String url, String method, String data,
      String contentLength) throws Utils.UserException {
    if (!isFirstRequest())
      doInsertDelay(getTimeElapsedSinceLastestRequest());
    StringBuilder sample = new StringBuilder();
    
    sample.append("			<sample use=\"replayHttp\" name=\"" + method + "\">" + EOL);
    sample.append("				<params>" + EOL);
    sample.append("					<param name=\"password\" value=\"\"></param>" + EOL);
    sample.append("					<param name=\"parameters\" value=\"" + queryStringParameters + "\"></param>" + EOL);
    if (method.equalsIgnoreCase("post"))
    {
    	sample.append("					<param name=\"bodyparameters\" value=\"" + bodyParameters + "\"></param>" + EOL);
    }
    sample.append("					<param name=\"realm\" value=\"\"></param>" + EOL);
    sample.append("					<param name=\"proxyport\" value=\"\"></param>" + EOL);
    sample.append("					<param name=\"proxyhost\" value=\"\"></param>" + EOL);
    sample.append("					<param name=\"cookies\" value=\"\"></param>" + EOL);
    sample.append("					<param name=\"proxypassword\" value=\"\"></param>" + EOL);
    sample.append("					<param name=\"hostauth\" value=\"\"></param>" + EOL);
    sample.append("					<param name=\"cookiepolicy\" value=\"\"></param>" + EOL);
    sample.append("					<param name=\"uri\" value=\"" + url + "\"></param>" + EOL);
    sample.append("					<param name=\"username\" value=\"\"></param>" + EOL);
    sample.append("					<param name=\"redirect\" value=\"\"></param>" + EOL);
    sample.append("					<param name=\"localaddress\" value=\"\"></param>" + EOL);
    sample.append("					<param name=\"proxylogin\" value=\"\"></param>" + EOL);
    sample.append("				</params>" + EOL);
    sample.append("			</sample>" + EOL);
    insertStmt(sample.toString());
	queryStringParameters = "";
	bodyParameters = "";
  }

  public void doAssertResponse(String respCode) {
  }

  public void doTidyCode(String url) {
  }

  public void doResponseForStdOut(String url) {
  }

  public void doResponseForFile() {
  }

  public void doEndTransaction() throws UserException {
  }

  public void doStartRecording() {
  }

  public void doStopRecording() {
  }

  public void doClose() {
  }

  public void setStruts(boolean struts) {
  }

  public void doHeaders(HeaderValue[] headers) {

  }

  private void wl(String s) {
    getScriptAdapter().append(s + EOL);
  }

  /**
   * Inserts a string into the script, with a newline appended.
   */
  private void insertStmt(String s) throws Utils.UserException {
    insert(s + EOL);
  }

  /**
   * Insert a delay with the timer plugin specified in the configuration file
   * 
   * @param delay
   * @throws Utils.UserException
   */
  private void doInsertDelay(long delay) throws Utils.UserException {
	StringBuilder timerStr = new StringBuilder();
	if (timer.equals(RANDOM_TIMER))
	{
	  int min = (int)(delay - delta);
	  if (min < 0) {
		min = 0;
	  }
	  if (randomDist.equals(RANDOM_TIMER_DIST_UNIFORM)) {
		timerStr.append("			<control use=\"replayTimer\" name=\"setUniform\">" + EOL);
		timerStr.append("				<params>" + EOL);
		timerStr.append("					<param name=\"min\" value=\"" + min + "\"/>" + EOL);
		timerStr.append("					<param name=\"max\" value=\"" + (delay + delta) + "\"/>" + EOL);
		timerStr.append("				</params>" + EOL);
		timerStr.append("			</control>" + EOL);
	  }
	  else if (randomDist.equals(RANDOM_TIMER_DIST_GAUSSIAN)) {
		timerStr.append("			<control use=\"replayTimer\" name=\"setGaussian\">" + EOL);
		timerStr.append("				<params>" + EOL);
		timerStr.append("					<param name=\"min\" value=\"" + min + "\"/>" + EOL);
		timerStr.append("					<param name=\"max\" value=\"" + (delay + delta) + "\"/>" + EOL);
		timerStr.append("					<param name=\"mean\" value=\"" + delay + "\"/>" + EOL);
		timerStr.append("					<param name=\"deviation\" value=\"" + deviation + "\"/>" + EOL);
		timerStr.append("				</params>" + EOL);
		timerStr.append("			</control>" + EOL);
	  }
	  else if (randomDist.equals(RANDOM_TIMER_DIST_POISSON)) {
		timerStr.append("			<control use=\"replayTimer\" name=\"setPoisson\">" + EOL);
		timerStr.append("				<params>" + EOL);
		timerStr.append("					<param name=\"unit\" value=\"" + unit + "\"/>" + EOL);
		timerStr.append("					<param name=\"parameter\" value=\"" + delay + "\"/>" + EOL);
		timerStr.append("				</params>" + EOL);
		timerStr.append("			</control>" + EOL);
	  }
	  else if (randomDist.equals(RANDOM_TIMER_DIST_NEGEXPO)) {
		timerStr.append("			<control use=\"replayTimer\" name=\"setNegativeExpo\">" + EOL);
		timerStr.append("				<params>" + EOL);
		timerStr.append("					<param name=\"min\" value=\"" + min + "\"/>" + EOL);
		timerStr.append("					<param name=\"mean\" value=\"" + delay + "\"/>" + EOL);
		timerStr.append("				</params>" + EOL);
		timerStr.append("			</control>" + EOL);
	  }
      timerStr.append("			<timer use=\"replayTimer\" name=\"sleep\"/>" + EOL);
	}
	else if (timer.equals(CONSTANT_TIMER)) {
      timerStr.append("			<timer use=\"replayTimer\" name=\"sleep\">" + EOL);
      timerStr.append("				<params>" + EOL);
      timerStr.append("					<param name=\"duration_arg\" value=\"" + delay + "\"></param>" + EOL);
      timerStr.append("				</params>" + EOL);
      timerStr.append("			</timer>" + EOL);
    }
	insertStmt(timerStr.toString());
  }

  public void doSetCharset(String cs) throws UserException {
  }
}
