package com.bitmechanic.maxq;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.python.util.PythonInterpreter;

import java.io.IOException;
import java.io.InputStream;
import java.util.*;
import org.apache.log4j.Logger;
import org.apache.log4j.Level;
import org.apache.log4j.Layout;
import org.apache.log4j.ConsoleAppender;
import org.apache.log4j.PatternLayout;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;

/**
 * This singleton class stores all config information for MAXQ.  Config data can
 * be set through the maxq.properties file, or the command line.
 * <p/>
 * Some people might think that this class is overkill, duplicating settings
 * that need not be stored, etc.  However those cynics should consider that
 * this class is referred to in unit tests, and would also enable someone
 * running MaxQ scripts from a JUnit test harness to modify options.  Otherwise
 * this would be impossible because the command-line would never be parsed.
 */
public class Config
{
  private static Config config = null;
  private static final Log log = LogFactory.getLog(Config.class);
  private static Properties props;

  private String pythonPath;
  private int port = 8090;
  private boolean debug;
  private boolean quiet;
  private String exist;
  private String replace;
  private List includeList;
  private List excludeList;
  private List includePatterns = new ArrayList();
  private List excludePatterns = new ArrayList();
  private String scriptArg;
  private ProxySettings proxySettings;

  static {
    Layout layout = new PatternLayout("%d [%t] %-5p %c - %m%n");
    Logger.getRootLogger().addAppender(new ConsoleAppender(layout, "System.err"));
  }
  
  private static String defDriverPkgName = "com.bitmechanic.maxq";
  private static String defValidatorPkgName = "com.bitmechanic.maxq";

  // Values that indicate a "true" boolean value.
  //  private String[] trueValues = {"yes", "true", "y", "t", "1"};
  //  private String[] falseValues = {"no", "false", "n", "f", "0"};

  /**
   * This class represents the settings for the proxy *THAT MAXQ USES*.
   * So, if set, maxq itself will proxy its requests.
   */
  public static class ProxySettings
  {
    String host;
    int port;

    public String getHost()
    {
      return host;
    }

    public void setHost(String host)
    {
      this.host = host;
    }

    public int getPort()
    {
      return port;
    }

    public void setPort(int port)
    {
      this.port = port;
    }
  }

  public static Config getConfig()
  {
    // if config is null, it means that the Config has not yet been
    // explicitly initialised, so the caller expects the 'default
    // behaviour' which is to load it from the classpath.
    if (config == null)
      initConfig();
    return config;
  }

  /**
   * Set up the properties file from the properties.file system
   * property, defaulting to maxq.properties if it exists.
   */
  public static void initConfig()
  {
    config = new Config();
    String maxqDir = System.getProperty("maxq.dir");
    String pathSep = System.getProperty("file.separator");

    // Load maxq.properties from the CLASSPATH
    String propertiesFileName = System.getProperty("properties.file", "maxq.properties");
    InputStream propertiesStream = Config.class.getClassLoader().getResourceAsStream(propertiesFileName);
    props = new Properties();
    if (propertiesStream == null) {
      // If we are here then log4j won't be able to find its configuration
      // either, and will not output errors to stderr :-(
      System.err.println("Could not find " + propertiesFileName + " in CLASSPATH.  Defaults will be used, and logging disabled.");
    } else {
      try {
        props.load(propertiesStream);
        log.debug("Generator: " + props.getProperty("generator.classnames"));
      } catch (IOException e) {
        // We do not want to absorb the exception without reporting
        // it so we convert it to what Java uses to report Errors.
        throw new ExceptionInInitializerError(e);
      }
    }

    // Add a default location to the pythonPath member variable
    config.addPythonPath(maxqDir + pathSep + "jython");
    // if the user has defined additional locations, add them to the same meber variable
    String path = config.getProperty("python.path");
    if (path != null)
      config.addPythonPath(path);

    // Load default port now so it can be overwritten by the command line.
    String portStr = props.getProperty("local.proxy.port");
    if (portStr != null)
      config.port = Integer.parseInt(portStr);
  }

  private Config()
  {
  }

  /**
   * Call this when all configuration options have been set.
   */
  public void completeInit() throws Utils.UserException
  {
    // Specify how much logging we want.  We set it for the root rather than
    // com.bitmechanic.maxq because otherwise the root will filter our
    // messages out.
    Level level = debug ? Level.ALL : quiet ? Level.WARN : Level.INFO;
    Logger.getRootLogger().setLevel(level);
 
    // Getting the Logger via the Apache interface here is a good idea
    // because it will fail if apache is using some other log class for
    // some reason.
    //    // Ask for less logging from HttpClient than from our own
    //    // logging.  This is particularly to disable annoying INFO messages.
    //    level = debug ? Level.INFO : Level.WARN;
    //    Log apacheLog = LogFactory.getLog("org.apache.commons");
    //    ((Jdk14Logger) apacheLog).getLogger().setLevel(level);
    // An alternative way to achieve the above.
    //System.setProperty("org.apache.commons.logging.Log", "org.apache.commons.logging.impl.SimpleLog");
    System.setProperty("org.apache.commons.logging.simplelog.defaultlog", "warn");

    // We always initialize the python interpreter because our
    // generator classes may call it.
    Properties ppref = new Properties();
    if (getPythonPath() != null)
      ppref.put("python.path", getPythonPath());
    PythonInterpreter.initialize(System.getProperties(), ppref, new String[0]);

    // Load proxy settings.
    String host = getProperty("remote.proxy.host");
    Integer port = getPropertyInt("remote.proxy.port");
    if ((host == null) != (port == null))
      throw new Utils.UserException("Not using proxy server.  You must set both remote.proxy.host and remote.proxy.port.");
    if (host != null && port != null) {
      proxySettings = new ProxySettings();
      proxySettings.host = host;
      proxySettings.port = port.intValue();
      log.info("Proxying requests via " + proxySettings.host + ":" + Integer.toString(proxySettings.port));
    }
  }

  /**
   * Returns null if the property is not found.
   */
  public String getProperty(String prop)
  {
    return (String) props.get(prop);
  }

  /**
   * Returns the string value of the property, or the given default if the
   * property is missing or empty.
   */
  public String getProperty(String prop, String def)
  {
    String val = getProperty(prop);
    return ((val == null) || val.equals("")) ? def : val;
  }

  public void setProperty(String property, String value)
  {
    props.setProperty(property, value);
  }

  /**
   * Returns the property as an integer.
   */
  public Integer getPropertyInt(String prop) throws Utils.UserException
  {
    String s = getProperty(prop);
    if (s == null)
      return null;
    try {
      return new Integer(s);
    } catch (NumberFormatException e) {
      throw new Utils.UserException(prop + " property must be an integer");
    }
  }

  public void addPythonPath(String path)
  {
    if (pythonPath == null)
      pythonPath = path;
    else
      pythonPath = pythonPath + System.getProperty("path.separator") + path;
  }

  public String getPythonPath()
  {
    return pythonPath;
  }

  public void setPort(int pt)
  {
    port = pt;
  }

  public int getPort()
  {
    return port;
  }

  public void setDebug(boolean d)
  {
    debug = d;
  }

  public boolean isDebug()
  {
    return debug;
  }

  public void setQuiet(boolean q)
  {
    quiet = q;
  }

  public boolean isQuiet()
  {
    return quiet;
  }

  public void setExist(String e)
  {
    exist = e;
  }

  public String getExist()
  {
    return exist;
  }

  public void setReplace(String e)
  {
    replace = e;
  }

  public String getReplace()
  {
    return replace;
  }

  public void setIncludePatterns(String[] pats)
  {
    includeList = Arrays.asList(pats);
    createPatterns(includeList, includePatterns);
  }

  public void setExcludePatterns(String[] pats)
  {
    excludeList = Arrays.asList(pats);
    createPatterns(excludeList, excludePatterns);
  }

  public List getIncludePatterns()
  {
    return includePatterns;
  }

  public List getExcludePatterns()
  {
    return excludePatterns;
  }

  public void setScriptArg(String arg)
  {
    scriptArg = arg;
  }

  public String getScriptArg()
  {
    return scriptArg;
  }

  public static String getDriverPkgName()
  {
    return System.getProperty("maxq.driverpkgname", defDriverPkgName);
  }

  public static String getValidatorPkgName()
  {
    return System.getProperty("maxq.validatorpkgname", defValidatorPkgName);
  }

  public static Log getTestLogger()
  {
    return LogFactory.getLog("com.bitmechanic.maxq.testrun");
  }

  public ProxySettings getProxySettings()
  {
    return proxySettings;
  }

  public void setProxySettings(ProxySettings proxySettings)
  {
    this.proxySettings = proxySettings;
  }

  private void createPatterns(List list, List pats)
  {
    Iterator iter = list.iterator();
    pats.clear();
    while (iter.hasNext()) {
      try {
        Pattern p = Pattern.compile((String) iter.next());
        pats.add(p);
      } catch (PatternSyntaxException e) {
        log.error(e);
      }
    }
  }
}
