package com.bitmechanic.maxq;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import java.io.*;
import java.net.*;

/**
 * One of these is created as a thread to handle each request and response
 * pair for ProxyServer.
 */
public class RequestHandler implements Runnable
{
  private static final Log log = LogFactory.getLog(RequestHandler.class);

  private Config config = Config.getConfig();

  private InputStream clientIn, serverIn;
  private OutputStream clientOut, serverOut;

  private HttpRequestHeader header;

  private ProxyServer proxyServer;
  private Socket clientSocket, serverSocket;
  private ByteArrayOutputStream buffer;
  private static Object mutex = new Object();

  RequestHandler(ProxyServer proxyServer, Socket s)
  {
    assert(s != null);
    clientSocket = s;
    buffer = new ByteArrayOutputStream();
    this.proxyServer = proxyServer;
  }

  /**
   * Gets the client information and establishes a connection with the
   * server.  This method is blocking.
   */
  private void initClientServerConnections(Socket s) throws Throwable
  {
    // We buffer the input stream because HttpRequestHeader must read
    // portions byte by byte.
    clientIn = new BufferedInputStream(s.getInputStream());
    clientOut = s.getOutputStream();
    try {
      header = new HttpRequestHeader(clientIn);
    } catch (IOException e) {
      // Browser has stopped writing its request halfway through!
      // That can happen if the user requests some other page so it
      // is not worth reporting except in passing.
      log.info("truncated request from browser: " + e.getMessage());
      throw new Utils.SilentException();
    }
    if (!header.url.startsWith("http"))
      throw new Utils.UserException("MaxQ only supports the HTTP protocol.");
    URL url = new URL(header.url);

    // Should we send the request via a proxy server?
    Config.ProxySettings proxy = config.getProxySettings();
    String host;
    int port;
    if (proxy != null) {
      host = proxy.host;
      port = proxy.port;
    } else {
      host = url.getHost();
      port = url.getPort();
    }

    // Open a connection to the remote server (or proxy server).
    if (port < 1)
      port = 80;
    try {
      serverSocket = new Socket(InetAddress.getByName(host), port);
    } catch (ConnectException e) {
      String msg = "Cannot connect to ";
      if (proxy != null)
        msg = msg + "proxy server ";
      msg = msg + host;
      if (port != 80)
        msg = msg + " on port " + Integer.toString(port);
      throw new Utils.UserException(msg + ".");
    }
    try {
      serverIn = serverSocket.getInputStream();
      serverOut = serverSocket.getOutputStream();
    } catch (Throwable e) {
      serverSocket.close();
      serverSocket = null;
      throw e;
    }
  }

  /**
   * a request header is of this format:
   * Request-Line   = Method SP Request-URI SP HTTP-Version CRLF
   * <p/>
   * this method assumes the second space delimited token is the request uri
   * and so it takes that token and strips the full url.
   * <p/>
   * this is done becuase with HTTP 1.0 the request header should contain
   * full URL information only if it is being sent to a proxy server, and
   * the proxy server is supposed to strip that information.
   */
  private String stripProxyInfoFromRequestHeader()
      throws MalformedURLException
  {
    //rewrite the uri
    String res = "";
    String origUrl = header.url;
    URL url = new URL(origUrl);
    header.url = url.getFile();
    res = header.toString();
    header.url = origUrl;
    return res;
  }

  /**
   * does the following <ol>
   * <li> get header info from client.
   * <li> strips all but relative address info from header
   * <li> sends that information to the rewritten header, along with all
   * the rest of the information to the server.
   * <p/>
   * <li> gets the server response
   * <li> passes that information back to the client.
   * <li> the server should close its connection. once that happens, the
   * proxy closes its connection.
   * </ol>
   */
  public void run()
  {
    try {
      try {
        final int BUF_SIZE;
        {
          int rs = clientSocket.getReceiveBufferSize();
          int ss = clientSocket.getSendBufferSize();
          BUF_SIZE = rs < ss ? ss : rs;
        }

        byte buf[] = new byte[BUF_SIZE];

        initClientServerConnections(clientSocket);

        //read the header and strip proxy info
        String headerStr;
        if (config.getProxySettings() == null)
          headerStr = stripProxyInfoFromRequestHeader();
        else
          headerStr = header.toString();
        log.trace("read request header");

        //send the byte information to the server
        {
          byte bytes[] = headerStr.getBytes();
          serverOut.write(bytes, 0, bytes.length);
        }
        log.trace("wrote request header");

        //next get the bytes for the rest of the request
        //and send them
        byte requestBody[];
        if (header.contentLength > 0) {
          buffer.reset();
          int len = 0;
          int num = 0;
          while (num < header.contentLength) {
            try {
              len = clientIn.read(buf, 0, buf.length);
            } catch (SocketException e) {
              // Browser has stopped writing its request halfway
              // through!  That can happen if the user requests
              // some other page so it is not worth reporting
              // except in passing.
              log.info("truncated request from browser: "
                      + e.getMessage());
              throw new Utils.SilentException();
            }
            if (len == 0)
              break;
            log.trace("read " + Integer.toString(len) + " bytes");
            serverOut.write(buf, 0, len);
            buffer.write(buf, 0, len);
            log.trace("wrote " + Integer.toString(len) + " bytes");
            num += len;
          }
          requestBody = buffer.toByteArray();
          log.trace("transferred rest of request body");
        } else {
          requestBody = new byte[0];
          log.trace("no request body");
        }

        // We want nothing more from the client because we do not handle
        // keep-alive.  (The lack of this caused a hang after the first
        // request on IE6.)  Perhaps we should be stripping
        // "Proxy-Connection: Keep-Alive" from the request header too.
        clientSocket.shutdownInput();

        // Similarly, we have nothing more for the server.  This will
        // ensure it feels that now is the time to get back to us.
        serverSocket.shutdownOutput();

        // The script will be screwed if a call to processRequest
        // is not followed by a call to processResponse.  This may
        // occur when there are multiple requests being handled
        // very close together.  This happens in practice
        // when redirections are encounted.  By using a mutex
        // we avoid this.  (This could be more thread-efficient if we
        // saved the request and response in separate buffers and
        // reported on them together in quick succession after we
        // had done all our TCP communication.)
        synchronized (mutex)
        {
          proxyServer.processRequest(header, requestBody);
          log.trace("processed request");

          //get the server reply
          //todo use a different header for response header
          buffer.reset();
          int len;
          while ((len = serverIn.read(buf, 0, buf.length)) > 0) {
            log.trace("read " + Integer.toString(len));
            try {
              clientOut.write(buf, 0, len);
            } catch (SocketException e) {
              // This will happen if the user cancels.  We do not
              // want it to interrupt the MaxQ GUI because it happens
              // frequently, particularly when the user click on
              // another link before the current page finishes loading
              log.info("browser stopped listening: "
                      + e.getMessage());
              throw new Utils.SilentException();
            }
            buffer.write(buf, 0, len);
          }
          log.trace("transferred response");
          //todo support for redirects via response code 302 etc...
          proxyServer.processResponse(header, buffer.toByteArray());
          log.trace("processed response");
        }
      } finally {
        // We used not to close serverSocket.  A comment said:
        // //the server should close its socket
        // //automatically
        // However they did not _appear_ to be closing, and this does
        // not surprise me because we are probably passing Keep-Alive
        // headers through to the server.  oliverbock
        if (serverSocket != null) {
          serverSocket.close();
          log.trace("closed server socket");
        }

        //close the client connection
        clientSocket.close();
        log.trace("closed client socket");
      }
    } catch (Throwable ex) {
      if (ex.getClass() != Utils.SilentException.class)
        SwingMain.reportOutsideException(ex);
    }
  }
}
