package com.bitmechanic.maxq.generator;

import com.bitmechanic.maxq.HeaderValue;
import com.bitmechanic.maxq.IScriptAdapter;
import com.bitmechanic.maxq.Param;
import com.bitmechanic.maxq.Utils;
import com.bitmechanic.maxq.Utils.UserException;

import org.python.core.PyString;

/**
 * This class generates scripts in the format required by the CompactTest
 * base class.
 * <p/>
 * IScriptGenerator is too finely grained for our needs so this class is
 * more complicated than it needs to be.
 * <p/>
 * This class improves upon JythonCodeGenerator as follows:
 * - It produces more concise scripts (thanks to a more sophisticated base
 * class).  The script will contain only _what_ was recorded, not _how_ to
 * test/playback.
 * - You can restart recording on an existing script.
 */
public class CompactGenerator extends AbstractCodeGenerator
{
  // Data collected from callbacks.  We output everything for each
  // fetch in one go in doEndTransaction().
  Param[] reqParams;
  String reqBaseUrl, reqMethod, respCode;
  //private boolean struts = false;
  
  public CompactGenerator(IScriptAdapter adapter)
  {
    super(adapter,
        new String[]{"^class\\s+(\\w+)",
              "(\\w+)\\('\\w+'\\).Run()",
              "\\w+\\('(\\w+)'\\).Run()"});
  }
  
  public void setStruts(boolean struts)
  {
    // Not used anywhere.
    //this.struts = struts;
  }
  
  public static String getGeneratorDescription()
  {
    return "Compact script";
  }

  public void doNew()
  {
    // Boilerplate.
    getScriptAdapter().append("# Generated by MaxQ [" + getClass().getName() + "]" + EOL +
            "from CompactTest import CompactTest" + EOL +
            EOL +
            "class " + getDefaultTestName() + "(CompactTest):" + EOL +
            "    # Recorded test actions." + EOL +
            "    def runTest(self):" + EOL +
            "        self.msg('Test started')" + EOL +
            EOL +
            // Keep the lines below in sync. with insertMarkerRE (superclass).
            "    # ^^^ Insert new recordings here.  (Do not remove this line.)" + EOL +
            EOL +
            EOL +
            "# Code to load and run the test" + EOL +
            "if __name__ == 'main':" + EOL +
            "    " + getTestName() + "('" + getTestName() + "').Run()" + EOL);
  }

  // All requests are put in one function so there is no need to do anything
  // when recording starts or stops.
  public void doStartRecording()
  {
  }

  public void doStopRecording()
  {
  }

  public String[] getValidFileExtensions()
  {
    return new String[]{".py"};
  }

  // These procedures accumulate data to be written out in a chunk later.
  public void doParameterList(Param[] params)
  {
    reqParams = params;
  }

  public void doCallUrl(String url, String method, String data,
                        String contentLength)
  {
    reqBaseUrl = url;
    reqMethod = method;
  }

  public void doAssertResponse(String code)
  {
    respCode = code;
  }

  // Finally we have all the information needed for a complete request.
  public void doEndTransaction() throws Utils.UserException
  {
    // Build the text for this request and response.
    String s = "        self." + reqMethod + "(";
    int len = s.length();
    s = s + quote(reqBaseUrl) + ", ";
    if (reqParams == null)
      s = s + "None";
    else {
      s = s + "[" + EOL;
      int i;
      String pad = new String();
      for (i = 0; i < len; i++)
        pad = pad + " ";
      for (i = 0; i < reqParams.length; i++) {
        s = s + pad + "(" + quote(reqParams[i].name) + ", "
            + quote(reqParams[i].value) + ")," + EOL;
      }
      s = s + pad + "]";
    }
    s = s + ", " + respCode + ")" + EOL;

    // Clear transaction storage variables.
    reqParams = null;
    reqBaseUrl = null;
    reqMethod = null;
    respCode = null;

    // Patch the changes into the script.
    insert(s);
  }

  public void doSetCharset(String cs) throws Utils.UserException
  {
    insert("   self.setCharset("+quote(cs)+")");
  }

  public void doHeaders(HeaderValue[] headers)
  {
    //To change body of implemented methods use File | Settings | File Templates.
  }

  // This generator does not support this stuff.
  public void doTestUrlMessage(String url)
  {
  }

  public void doSetData(String data)
  {
  }

  public void doTidyCode(String url)
  {
  }

  public void doResponseForStdOut(String url)
  {
  }

  public void doResponseForFile()
  {
  }

  /**
   * Quotes the string using Python conventions.
   */
  private String quote(String s)
  {
    return new PyString(s).__repr__().toString();
  }

  public void doQueryStringParameterList(Param[] params) throws UserException {
	// TODO Auto-generated method stub	
  }

  public void doBodyParameterList(Param[] params) throws UserException {
	// TODO Auto-generated method stub
  }
}
