package com.bitmechanic.maxq.generator;

import com.bitmechanic.maxq.ProxyServer;

/**
 * All classes that generate MaxQ scripts must implement this interface.
 * Then they may be listed in generator.classnames in maxq.properties.
 * <p/>
 * Notes:
 * - In practice derived classes are descendents of @AsbtractCodeGenerator.
 * - All generators must put their classname somewhere in the script.
 * This is used to recognise which generator to use when a script is
 * reloaded and recording is resumed.
 *
 * @author Robert Dawson robert@rojotek.com
 * @author Wayne Hearn  wayne_hearn@dell.com
 * @see GeneratorFactory
 */
public interface IScriptGenerator extends ProxyServer.Observer
{
  public static String EOL = System.getProperty("line.separator");

  /**
   * Returns the description of this generator class for display in the GUI.
   * This cannot be part of the interface because the control freaks in
   * charge of Java say so.  We use the Reflection API to find it, and
   * do not mind if it is missing.
   */
  //public static String getGeneratorDescription();

  /**
   * Returns  an array of valid file extensions for the type of code
   * generator in use.
   *
   * @return array of valid file extentions
   */
  public String[] getValidFileExtensions();

  /**
   * Parses the name of the test out of the script.
   *
   * @return test name if test has been loaded or saved.  Empty otherwise.
   */
  public String parseTestName();

  /**
   * Called before any processRequest() calls.
   */
  public void doStartRecording();

  /**
   * Called after the last processResponse() call, but may be followed by
   * another doStartRecording() to start the process over again.
   */
  public void doStopRecording();

  /**
   * Called when a new script needs to be begun.  This will not be called
   * when an existing script is loaded from file.
   * <p/>
   * The script should output its full class name during this call to ensure
   * that its generator can be recognised if it is saved and reloaded.
   */
  public void doNew();

  /**
   * Notifies the generator that the script is about to be saved.  This gives
   * it the opportunity to patch the script to include the new filename.
   */
  public void doSave(String path, String fileName);

  /**
   * Called after a new script has been loaded from file.
   */
  public void doLoad();

  /**
   * Called when a generator is no longer needed.
   */
  public void close();
}
