// 
// Copyright (c) 2005, Caltha - Gajda, Krzewski, Mach, Potempski Sp.J. 
// All rights reserved. 
// 
// Redistribution and use in source and binary forms, with or without modification,  
// are permitted provided that the following conditions are met: 
//  
// * Redistributions of source code must retain the above copyright notice,  
//	 this list of conditions and the following disclaimer. 
// * Redistributions in binary form must reproduce the above copyright notice,  
//	 this list of conditions and the following disclaimer in the documentation  
//	 and/or other materials provided with the distribution. 
// * Neither the name of the Caltha - Gajda, Krzewski, Mach, Potempski Sp.J.  
//	 nor the names of its contributors may be used to endorse or promote products  
//	 derived from this software without specific prior written permission. 
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"  
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED  
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
// IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,  
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  
// BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
// OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,  
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE  
// POSSIBILITY OF SUCH DAMAGE. 
// 
package com.bitmechanic.maxq.generator;

import java.io.ByteArrayOutputStream;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.bitmechanic.maxq.Config;
import com.bitmechanic.maxq.HeaderValue;
import com.bitmechanic.maxq.HttpRequestHeader;
import com.bitmechanic.maxq.Param;
import com.bitmechanic.maxq.Utils;

/**
 * Describes browser / application interchange composed of one request and it's response.
 * 
 * <p>Caveats: at this point, only few request headers are handled, response headers are not
 * handled at all, request / repsonse cookies are not handled, and upload parameters are not 
 * processed correctly.</p>
 * 
 * <p>Character encodings: all byte -> String conversions are using platform default encoding. 
 * This is going to work in the following situations:
 * <ul>
 *   <li>your application is using only 7bit USASCII characters.</li>
 *   <li>your application is using UTF-8 encoding, and so is the JVM where MaxQ is running.
 *   pass <code>-Dfile.encoding=UTF-8</code> if your OS is using something different.</li>
 * </ul>
 * In all other cases, expect chaos and misery.</p>
 * 
 * @author <a href="mailto:rafal@caltha.pl">Rafal Krzewski</a>
 * @author MaxQ community
 * @version $Id: Interchange.java,v 1.1 2005/03/17 10:54:38 rkrzewsk Exp $
 */
public class Interchange {

    /** previous interchange, may be null */
    private final Interchange previous;

    /** the log. */
    private final Log log = LogFactory.getLog(getClass());

    /** MaxQ config. */
    private final Config config;

    /** This interchange is to be ignored. */
    private boolean ignored;

    // request ----------------------------------------------------------------------------------

    /** request url without query string and fragment */
    private String baseUrl;

    /** complete request url */
    private String url;

    /** request method. */
    private String method;

    /**
     * Map<String,String> (name, value)
     */
    private Map requestHeaders = new HashMap();

    /**
     * Map<String,String> (name, value)
     */
    private Map requestCookies = new HashMap();

    /** Map<String,String> (name,value) */
    private Map getParameters = new HashMap();

    /** Map<String,String> (name, value) */
    private Map postParameters = new HashMap();

    /** Map<String,String> (name, value) concat of getParameters and postParameters */
    private Map allParameters = new HashMap();

    /** Map<String,String> (name, path) */
    private Map uploadParamters = new HashMap();

    // -- response ------------------------------------------------------------------------------

    /** response status code. */
    private String statusCode;

    /** response content type. */
    private String contentType;

    /** response body. */
    private String responseBody;

    /** Map<String,String> (name, value) */
    private Map responseHeaders = new HashMap();

    /**
     * Map<String,String> (name, value)
     */
    private Map responseCookies = new HashMap();

    // -- accessors -----------------------------------------------------------------------------

    /**
     * Returns the previous.
     * 
     * @return the previous.
     */
    public Interchange getPrevious() {
        return previous;
    }

    /**
     * Returns the ignored.
     * 
     * @return the ignored.
     */
    public boolean isIgnored() {
        return ignored;
    }

    /**
     * Returns the base url.
     * 
     * @return the base url.
     */
    public String getBaseUrl() {
        return baseUrl;
    }

    /**
     * Returns the url.
     * 
     * @return the url.
     */
    public String getUrl() {
        return url;
    }

    /**
     * Returns the method.
     * 
     * @return the method.
     */
    public String getMethod() {
        return method;
    }

    /**
     * Returns the requestHeaders.
     * 
     * @return the requestHeaders.
     */
    public Map getRequestHeaders() {
        return requestHeaders;
    }

    /**
     * Returns the requestCookies.
     * 
     * @return the requestCookies.
     */
    public Map getRequestCookies() {
        return requestCookies;
    }

    /**
     * Returns the getParameters.
     * 
     * @return the getParameters.
     */
    public Map getGetParameters() {
        return getParameters;
    }

    /**
     * Returns the postParameters.
     * 
     * @return the postParameters.
     */
    public Map getPostParameters() {
        return postParameters;
    }

    /**
     * Returns the allParamters.
     * 
     * @return the allParamters.
     */
    public Map getAllParameters() {
        return allParameters;
    }

    /**
     * Returns the uploadParamters.
     * 
     * @return the uploadParamters.
     */
    public Map getUploadParamters() {
        return uploadParamters;
    }

    /**
     * Returns the statusCode.
     * 
     * @return the statusCode.
     */
    public String getStatusCode() {
        return statusCode;
    }

    /**
     * Returns the contentType.
     * 
     * @return the contentType.
     */
    public String getContentType() {
        return contentType;
    }

    /**
     * Returns the responseBody.
     * 
     * @return the responseBody.
     */
    public String getResponseBody() {
        return responseBody;
    }

    /**
     * Returns the responseHeaders.
     * 
     * @return the responseHeaders.
     */
    public Map getResponseHeaders() {
        return responseHeaders;
    }

    /**
     * Returns the responseCookies.
     * 
     * @return the responseCookies.
     */
    public Map getResponseCookies() {
        return responseCookies;
    }

    // -- initialization ------------------------------------------------------------------------

    /**
     * Creates a new Interaction instance.
     * 
     * @param previous the previous interaction.
     */
    public Interchange(Config config, Interchange previous) {
        this.config = config;
        this.previous = previous;
    }

    // -- data capture --------------------------------------------------------------------------

    /**
     * Capture the request data.
     * 
     * @param header request header.
     * @param body request body.
     */
    public void processRequest(HttpRequestHeader header, byte[] body) {
        if (header.method.toLowerCase().equals("get") && urlIgnored(header.url)) {
            log.debug("Ignoring GET request: \"" + header.url + "\"");
            ignored = true;
            return;
        }

        url = replaceURL(header.url);

        int queryPos = header.url.indexOf("?");
        if (queryPos != -1 && (header.url.length() > queryPos + 1))
            baseUrl = replaceURL(header.url.substring(0, queryPos));
        else
            baseUrl = url;

        method = header.method;

        // TODO silly... we should parse the headers, they're actually present in body
        if (header.contentType.length() > 0)
            requestHeaders.put(HeaderValue.ENTITY_CONTENT_TYPE,
                    header.contentType);
        if (header.accept.length() > 0)
            requestHeaders.put(HeaderValue.REQUEST_ACCEPT, header.accept);
        if (header.referer.length() > 0)
            requestHeaders.put(HeaderValue.REQUEST_REFERER, header.referer);
        if (header.pragmaNoCache)
            requestHeaders.put(HeaderValue.GENERAL_CACHE_CONTROL, "no-cache");

        // TODO parse request cookies from the headers

        boolean isMultiPartRequest = header.contentType
                .startsWith("multipart/form-data");

        if (queryPos != -1 && (header.url.length() > queryPos + 1))
            setParameters(getParameters, header.url.substring(queryPos + 1));

        if (body.length > 0) {
            if (isMultiPartRequest)
                setMultipartParameters(header, body);
            else
                setParameters(postParameters, new String(body)); // TODO encoding
        }
        
        allParameters.putAll(getParameters);
        allParameters.putAll(postParameters);
        // this is a bit weird but browser duplicates some GET parameters in POST message body...
        postParameters.keySet().removeAll(getParameters.keySet());
    }

    public void processResposne(HttpRequestHeader header, byte[] body) {
        responseBody = new String(body); // TODO encoding

        statusCode = parseResponseCode(responseBody);

        contentType = parseContentType(responseBody);

        // TODO parse response headers

        // TODO response cookies
    }

    // -- implementation ------------------------------------------------------------------------

    /**
     * Check if the request should be ignored due to URL exclusion.
     * 
     * @param str the URL.
     * @return <code>true</code> if the request should be ignored.
     */
    private boolean urlIgnored(String str) {
        // If "include" patterns are given, then a match must occur to
        // not-ignore the request.
        // If "exclude" patterns are given, then a match must NOT occur to
        // not-ignore the request.
        // Even if a match occurs with "include" patterns, an "exclude" pattern
        // must still NOT match to not-ignore the request.
        boolean ignore = false;
        // Iterator iter;
        List incPats = config.getIncludePatterns();
        List excPats = config.getExcludePatterns();
        if (!ignore && incPats.size() != 0) {
            log.debug("urlIgnored: checking for include matches");
            ignore = checkMatch(incPats.iterator(), str, false, true);
        }
        if (!ignore && excPats.size() != 0) {
            log.debug("urlIgnored: checking for exclude matches");
            ignore = checkMatch(excPats.iterator(), str, true, ignore);
        }
        return ignore;
    }

    /**
     * Matches a String agains a series of Pattern objects.
     * 
     * @param iter iterator over a collection of Pattern objects.
     * @param str the string to be checked.
     * @param res value to return if a match is found.
     * @param def value to return if no match is found.
     * @return value of res if a match is found, value of def otherwise.
     */
    private boolean checkMatch(Iterator iter, String str, boolean res,
            boolean def) {
        while (iter.hasNext()) {
            Matcher m = ((Pattern) iter.next()).matcher(str);
            log.debug("checkMatch: for \"" + str + "\" with pattern: \""
                    + m.pattern().pattern() + "\"");
            if (m.find()) {
                log.debug("checkMatch: Found match, returning: " + res);
                return res;
            }
        }
        log.debug("checkMatch: Didn't find match, returning: " + def);
        return def;
    }

    /**
     * URL-decode a String.
     * 
     * @author MaxQ commnunity.
     * @param s the string.
     * @return the URL-decoded string.
     */
    private String urlDecode(String s) {
        if (s.indexOf("+") == -1 && s.indexOf("%") == -1) {
            return s;
        }
        ByteArrayOutputStream out = new ByteArrayOutputStream(s.length());
        for (int count = 0; count < s.length(); count++) {
            if (s.charAt(count) == '%') {
                count++;
                int a = Character.digit(s.charAt(count++), 16);
                a = a << 4;
                int b = Character.digit(s.charAt(count), 16);
                // prefix ' (ascii 39) or " (ascii 132) with '\' (ascii 92)
                if (a + b == 39 || a + b == 132)
                    out.write(92);
                out.write(a + b);
            } else {
                if (s.charAt(count) == '+')
                    out.write(' ');
                else
                    out.write(s.charAt(count));
            }
        }

        return (out.toString());
    }

    /**
     * Replace the url according to MaxQ configuration.
     * 
     * @param url the URL
     * @return the replaced URL.
     */
    private String replaceURL(String url) {
        if (config.getExist() == null || config.getReplace() == null) {
            return url;
        } else {
            int pos = url.indexOf(config.getExist());
            if (pos == -1) {
                return url;
            } else {
                String str = url.substring(0, pos) + config.getReplace()
                        + url.substring(pos + config.getExist().length());
                return str;
            }
        }
    }

    /**
     * Parse parameters string and add items to a map.
     * 
     * @author MaxQ community
     * @param parameters the map to put parameters into.
     * @param str parameter string.
     */
    private void setParameters(Map parameters, String str) {
        List params = Utils.getParams(str);
        Iterator iter = params.iterator();
        while (iter.hasNext()) {
            Param param = (Param) iter.next();
            parameters.put(urlDecode(param.name), urlDecode(param.value));
        }
    }

    /**
     * Picks up parameters from multipart/form-data request body. Note that I have better things to
     * do than read the HTTP RFC so I'm only implementing what I can see is necessary for accepting
     * form data based on examples sent by my browser. oliverbock.
     * 
     * @author MaxQ community
     * @param header the response header.
     * @param strarray response body.
     */
    private void setMultipartParameters(HttpRequestHeader header,
            byte[] strarray) {
        // Extract the boundary between parts.
        String str = new String(strarray); // TODO encoding
        int begin = header.contentType.indexOf("boundary=");
        int end = header.contentType.indexOf("; ", begin);
        if (end == -1)
            end = header.contentType.length();
        String boundary = header.contentType.substring(begin + 9, end);

        // Parse a parameter from each part.
        String[] parts = str.split("--" + boundary);
        assert parts[0].equals(""); // Separator begins data.
        int i;
        Pattern re = Pattern
                .compile(
                        "\r\nContent-Disposition: form-data; name=\"([^\"]+)\"[^\r\n]*\r\n\r\n(.*)\r\n",
                        Pattern.DOTALL);
        for (i = 1; i < parts.length - 1; i++) {
            Matcher m = re.matcher(parts[i]);
            boolean ok = m.matches();
            assert ok;
            postParameters.put(m.group(1), m.group(2));
        }

        // TODO pick up file uploads (name="..."; filename="...")

        // Observed terminator.
        assert parts[parts.length - 1].equals("--\r\n");
    }

    /**
     * Parser response status code from the response
     * 
     * @author MaxQ community
     * @param response the response body as a String.
     * @return response status code from the response
     */
    private String parseResponseCode(String response) {
        // parse the HTTP response code
        String respCode = null;
        int pos = response.indexOf(" ");
        if (pos != -1) {
            int end = response.indexOf(" ", pos + 1);

            // check for newline too
            int end2 = response.indexOf("\n", pos + 1);
            if (end2 != -1 && end2 < end)
                end = end2;
            if (end != -1)
                respCode = response.substring(pos + 1, end).trim();
        }

        log.debug("HTTP response code: " + respCode);
        return respCode;
    }

    /**
     * Parse content type from the response
     * 
     * @author MaxQ community
     * @param response response body as a String
     * @return the content type.
     */
    private String parseContentType(String response) {
        String contentType = null;

        int pos = response.indexOf("Content-Type:");
        if (pos > -1) {
            pos += 14; // okay we found 'content-type:', advance 14 char positions
            int end = response.indexOf("\r\n", pos);
            int end2 = response.indexOf(";", pos);
            if (end2 > -1 && end2 < end)
                end = end2;
            if (end > -1) {
                contentType = response.substring(pos, end).trim();
            }
            log.debug("  Content-Type: " + contentType);
        } else {
            log.debug("  No content-type header!  First few lines:");
            StringTokenizer st = new StringTokenizer(response, "\n");
            int i = 0;
            while (st.hasMoreTokens() && i < 5) {
                log.debug(st.nextToken());
                i++;
            }
        }
        return contentType;
    }

}