// 
// Copyright (c) 2005, Caltha - Gajda, Krzewski, Mach, Potempski Sp.J. 
// All rights reserved. 
// 
// Redistribution and use in source and binary forms, with or without modification,  
// are permitted provided that the following conditions are met: 
//  
// * Redistributions of source code must retain the above copyright notice,  
//	 this list of conditions and the following disclaimer. 
// * Redistributions in binary form must reproduce the above copyright notice,  
//	 this list of conditions and the following disclaimer in the documentation  
//	 and/or other materials provided with the distribution. 
// * Neither the name of the Caltha - Gajda, Krzewski, Mach, Potempski Sp.J.  
//	 nor the names of its contributors may be used to endorse or promote products  
//	 derived from this software without specific prior written permission. 
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"  
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED  
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
// IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,  
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  
// BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
// OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,  
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE  
// POSSIBILITY OF SUCH DAMAGE. 
// 
package com.bitmechanic.maxq.generator;

import java.io.IOException;
import java.io.InputStream;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.velocity.Template;
import org.apache.velocity.VelocityContext;
import org.apache.velocity.exception.MethodInvocationException;
import org.apache.velocity.exception.ParseErrorException;
import org.apache.velocity.exception.ResourceNotFoundException;
import org.apache.velocity.runtime.RuntimeInstance;

import com.bitmechanic.maxq.Config;
import com.bitmechanic.maxq.HttpRequestHeader;
import com.bitmechanic.maxq.IScriptAdapter;

/**
 * Template based code generator.
 * 
 * @author <a href="mailto:rafal@caltha.pl">Rafal Krzewski</a>
 * @author MaxQ community
 * @version $Id: TemplateCodeGenerator.java,v 1.2 2005/05/04 12:46:17 rkrzewsk Exp $
 */
public class TemplateCodeGenerator implements IScriptGenerator {

    /** Default template lookup paths. */
    private static final String DEFAULT_TEMPLATE_PATHS = 
        "templates:com/bitmechanic/maxq/generator/templates";

    /** Default template theme. */
    private static final String DEFAULT_THEME = "jWebUnit";
    
    /**
     * File name extensions supported by the generator.
     */
    private final String[] validFileExtensions;

    /**
     * Marker of the interchange code insertion in the script.
     */
    private final Pattern insertionMarkerPattern;

    /**
     * Marker of test session/scenarion header.
     */
    private final Pattern scenarioIdPattern;

    /**
     * MaxQ config object.
     */
    private final Config config = Config.getConfig();

    /**
     * Adapter of the script the generator operates on.
     */
    private final IScriptAdapter script;

    /**
     * The logger.
     */
    private final Log log = LogFactory.getLog(getClass());

    /**
     * Velocity engine instance.
     */
    private final RuntimeInstance velocityRuntime;

    /**
     * Template logic object.
     * 
     * @see TemplateLogic
     */
    private final Object templateLogic;

    /**
     * Template lookup directories under classpath (should not contain leading nor trailing 
     * slashes).
     */
    private final String[] templatePaths;
    
    /** Template theme name. */
    private final String theme;

    /**
     * Patterns for updating test name when it is changed during save operation.
     */
    // List<Pattern>
    private final List testNamePatterns = new ArrayList();

    /**
     * The current interchange.
     */
    private Interchange interchange = null;

    /**
     * Most recent non-ingnored interchange.
     */
    private Interchange lastNotIgnored = null;

    /**
     * Number of scenarios present in the script.
     */
    private int scenarios = 0;

    /**
     * Creates new TempateCodeGenerator instance.
     * 
     * @param script the script adapter.
     */
    public TemplateCodeGenerator(IScriptAdapter script) {
        this.script = script;

        // initialize Velocity
        InputStream is = getClass().getClassLoader().getResourceAsStream(
                "velocity.properties");
        if (is == null) {
            throw new RuntimeException(
                    "velocity.properties not found in the classpath");
        }
        Properties props = new Properties();
        try {
            props.load(is);
        } catch (IOException e) {
            throw new RuntimeException("failed to load velocity.properties", e);
        }
        velocityRuntime = new RuntimeInstance();
        try {
            velocityRuntime.init(props);
        } catch (Exception e) {
            throw new RuntimeException("failed to initialize velocity", e);
        }
        
        templatePaths = Pattern.compile(":").split(config.getProperty("template.paths", 
            DEFAULT_TEMPLATE_PATHS));

        // get user configured theme. This should chosen through UI in the future.
        theme = config.getProperty("template.theme", DEFAULT_THEME);

        // let the Test template set variables in the context
        VelocityContext context = new VelocityContext();
        context.put("config", config);
        try {
            merge(context, "Test");
        } catch (Exception e) {
            throw new RuntimeException("falied to parse configuration template", e);
        }
        
        // file extensions
        validFileExtensions = (String[])((List)context.get("validFileExtensions")).
            toArray(new String[1]);

        // prepare paterns
        try {
            insertionMarkerPattern = Pattern.compile((String)context.get("insertionMarkerPattern"), 
                Pattern.MULTILINE);
            
            List testNamePatternList = (List) context.get("testNamePatterns");
            if (testNamePatternList != null) {
                for (Iterator i = testNamePatternList.iterator(); i.hasNext();) {
                    testNamePatterns.add(Pattern.compile((String) i.next(),
                            Pattern.MULTILINE));
                }
            }
            
            scenarioIdPattern = Pattern.compile((String)context.get("scenarioIdPattern"), 
                Pattern.MULTILINE);
        } catch (Exception e) {
            throw new RuntimeException("failed to initilize patterns", e);
        }
        if (testNamePatterns.isEmpty()) {
            throw new RuntimeException(
                    "namePatterns variable not defined in the test template");
        }

        // prepare templateLogic object
        String templateLogicClass = (String) context.get("logicClass");
        if (templateLogicClass != null) {
            try {
                Class cl = Class.forName(templateLogicClass);
                templateLogic = cl.newInstance();
            } catch (Exception e) {
                throw new RuntimeException(
                        "failed to instantiate template.logic class", e);
            }
        } else {
            templateLogic = null;
        }
    }

    /**
     * Returns brief description of the generator.
     * 
     * @return brief description of the generator.
     */
    public static String getGeneratorDescription() {
        return "Template based script";
    }

    /**
     * {@inheritDoc}
     */
    public String[] getValidFileExtensions() {
        return validFileExtensions;
    }

    /**
     * {@inheritDoc}
     */
    public String parseTestName() {
        Matcher m = ((Pattern) testNamePatterns.get(0)).matcher(script.getText());
        if (!m.find()) {
            throw new IllegalArgumentException("Regular expression (\""
                    + ((Pattern) testNamePatterns.get(0)).pattern()
                    + "\") does not match the contents of the script");
        }
        return m.group(1);
    }

    /**
     * {@inheritDoc}
     */
    public void doStartRecording() {
        scenarios++;
        if (scenarios > 1) {
            // close previous scenario and start another
            VelocityContext context = new VelocityContext();
            context.put("scenarioId", new Integer(scenarios));
            if (templateLogic != null)
                context.put("logic", templateLogic);
            context.put("config", config);
            try {
                insert(merge(context, "NewScenario"));
            } catch (Exception e) {
                throw new RuntimeException(e);
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    public void doStopRecording() {
        // nothing to do
    }

    /**
     * {@inheritDoc}
     */
    public void doNew() {
        try {
            VelocityContext context = new VelocityContext();
            context.put("config", config);
            script.setText(merge(context, "Test"));
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * {@inheritDoc}
     */
    public void doSave(String path, String name) {
        int dotPos = name.indexOf(".");
        if (dotPos > -1)
            name = name.substring(0, dotPos);
        String oldName = null;
        // Update the name of the test in the file because it may have changed.
        for (int i = 0; i < testNamePatterns.size(); i++) {
            Matcher m = ((Pattern) testNamePatterns.get(i)).matcher(script
                    .getText());
            if (!m.find())
                throw new IllegalArgumentException("Regular expression (\""
                        + ((Pattern) testNamePatterns.get(i)).pattern()
                        + "\") does not match the contents of the script");
            // replace all occurances
            do {
                if(oldName == null)
                    oldName = m.group(1);
                else {
                    if(m.group(1).equals(oldName))
                        throw new IllegalArgumentException("Inconsistency in test name detection " +
                            oldName + " vs " + m.group(1));
                }
                script.replace(name, m.start(1), m.end(1));
            } while(m.find());            
        }
    }

    /**
     * {@inheritDoc}
     */
    public void doLoad() {
        // count the existing scenarios
        Matcher match = scenarioIdPattern.matcher(script.getText());
        int lastIndex = 0;
        while (match.find(lastIndex)) {
            lastIndex = match.end();
            int id = Integer.parseInt(match.group(1));
            if(id > scenarios)
                scenarios = id;
        }
    }

    /**
     * {@inheritDoc}
     */
    public void close() {
        // nothing to do
    }

    /**
     * {@inheritDoc}
     */
    public void processRequest(HttpRequestHeader header, byte[] body)
            throws Exception {
        interchange = new Interchange(config, lastNotIgnored);
        interchange.processRequest(header, body);
        log.debug("process request " + header.url);
    }

    /**
     * {@inheritDoc}
     */
    public void processResponse(HttpRequestHeader header, byte[] body)
            throws Exception {
        interchange.processResposne(header, body);
        log.debug("process response " + header.url);
        if (!interchange.isIgnored()) {
            lastNotIgnored = interchange;
            emitInterchange();
        }
    }

    // implementation ---------------------------------------------------------------------------

    /**
     * Emit a captured interchange to the script.
     */
    private void emitInterchange() {
        VelocityContext context = new VelocityContext();
        context.put("interchange", interchange);
        if (templateLogic != null)
            context.put("logic", templateLogic);
        context.put("config", config);
        try {
            insert(merge(context, "Interchange"));
        } catch(MethodInvocationException e) {
            throw new RuntimeException("template rendering error", e.getWrappedThrowable());
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * Merge a Velocity template.
     * 
     * <p>Templates are looked up in "templates/" directory in the classpath, then in
     * "com/bitmechanic/max/generator/templates/" in the classpath. This way users may
     * override templates provided by MaxQ or write their own.</p>
     * 
     * @param context the template context.
     * @param kind kind of the template.
     * @return merged template contetns.
     * @throws Exception if the rendering fails.
     */
    private String merge(VelocityContext context, String kind) throws Exception {
        Template template = getTemplate(theme + kind + ".vm");
        StringWriter writer = new StringWriter();
        template.merge(context, writer);
        return writer.getBuffer().toString();
    }

    /**
     * Attempts to load template from one of the configured directories.
     * 
     * @param name template name.
     * @return the template.
     * @throws ResourceNotFoundException if the template is not found.
     * @throws ParseErrorException if the template contains lexical errors.
     * @throws Exception if unspecified problem occurs.
     */
    private Template getTemplate(String name) throws ResourceNotFoundException,
            ParseErrorException, Exception {
        loop: for(int i = 0; i < templatePaths.length; i++) {
            try {
                return velocityRuntime.getTemplate(templatePaths[i] + "/" + name);
            } catch(ResourceNotFoundException e) {
                continue loop;
            }
        }
        throw new ResourceNotFoundException("template "+name+" not found in "+templatePaths);
    }

    /**
     * Insert content into the script before the match location of {@link #insertionMarkerPattern}.
     * 
     * @param content
     */
    private void insert(String content) {
        Matcher match = insertionMarkerPattern.matcher(script.getText());
        if (!match.find()) {
            throw new RuntimeException("marker not found");
        }
        script.insert(content, match.start());
    }
}
