// 
// Copyright (c) 2005, Caltha - Gajda, Krzewski, Mach, Potempski Sp.J. 
// All rights reserved. 
// 
// Redistribution and use in source and binary forms, with or without modification,  
// are permitted provided that the following conditions are met: 
//  
// * Redistributions of source code must retain the above copyright notice,  
//	 this list of conditions and the following disclaimer. 
// * Redistributions in binary form must reproduce the above copyright notice,  
//	 this list of conditions and the following disclaimer in the documentation  
//	 and/or other materials provided with the distribution. 
// * Neither the name of the Caltha - Gajda, Krzewski, Mach, Potempski Sp.J.  
//	 nor the names of its contributors may be used to endorse or promote products  
//	 derived from this software without specific prior written permission. 
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"  
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED  
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
// IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,  
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  
// BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
// OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,  
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE  
// POSSIBILITY OF SUCH DAMAGE. 
// 
package com.bitmechanic.maxq.generator;

import com.bitmechanic.maxq.Config;

/**
 * A helper object that can be used in the test templates.
 * 
 * <p>The {@link TemplateCodeGenerator} renders the <em>theme</em>Test.vm template, evaluating
 * the <code>#set()</code> expressions contained in there. If <code>$templateLogic</code> variable 
 * is set, it's value is used as a name of the class to be instantiated. A single instance of the 
 * class is associated with a <code>TemplateCodeGenerator</code> throughout its lifetime, therefore 
 * it is strongly recommended for the logic class to be stateless. The instance will be available 
 * during rendering <em>theme</em>NewScenario.vm and <em>theme</em>Interchange.vm templates in the 
 * <code>$logic</code> variable.
 * </p>
 * 
 * <p>No interface requirements are imposed on the class selected using <code>$templateLogic</code>
 * variable.
 *
 * @author <a href="mailto:rafal@caltha.pl">Rafal Krzewski</a>
 * @version $Id: TemplateLogic.java,v 1.1 2005/03/17 10:54:38 rkrzewsk Exp $
 */
public class TemplateLogic {

    /**
     * MaxQ configuration.
     */
    private final Config config = Config.getConfig();
    
    /**
     * The base URL.
     */
    private final String baseUrl;

    /**
     * Creates new TemplateLogic instance.
     */
    public TemplateLogic() {
        String url = config.getProperty("template.baseUrl", "http://localhost:8080");
        baseUrl = url.endsWith("/") ? url.substring(0, url.length()-1) : url;
    }

    // -- URLs ----------------------------------------------------------------------------------
    
    /**
     * Returns the applications base URL.
     * 
     * @return the applications base URL.
     */
    public String getBaseUrl() {
        return baseUrl;
    }

    /**
     * Returns an URL relative to the configured base URL.
     * 
     * @param url the URL.
     * @return an URL relative to the configured base URL.
     */
    public String getRelativeUrl(String url) {
        if (url.startsWith(baseUrl)) {
            if(baseUrl.equals(url)) 
                return "/";
            else
                return url.substring(baseUrl.length());
        } else {
            return null;
        }
    }

    /**
     * Escape a form parameter value in a way suitable for a String literal.
     *  
     * <p>" and \ characters will be escaped with a leading \.</p>
     * 
     * @param param parameter value.
     * @return escaped parameter value.
     */
    public String escapeParam(String param) {
        StringBuffer buff = new StringBuffer();
        for(int i = 0; i < param.length(); i++) {
            switch(param.charAt(i)) {
            case '"':
                buff.append("\\\"");
                break;
            case '\\':
                buff.append("\\\\");
                break;
            default:
                buff.append(param.charAt(i));
            }
        }
        return buff.toString();
    }

    /**
     * Returns the MaxQ config.
     * 
     * @return the MaxQ config.
     */
    protected Config getConfig()
    {
        return config;
    }
}
