from CompactTest import CompactTest
from com.bitmechanic.maxq import HttpTestCase
import os, os.path, re, difflib, types
import diff  # Aaron Swartz's HTML diff library

# Extends the standard test class to also record the response data we receive.
#
# Prerequisites:
# This class relies on Jython libraries that are not included wih the base
# MaxQ distribution.  See http://maxq.tigris.org/docs/edit.html
#
# Usage:
# - If no --script-arg (-g) is provided then the script will be run but the
#   responses will not be checked.  (Response codes will be checked, just the
#   content will be ignored.)
# - If --script-arg is "record" then the responses will be saved in a
#   subdirectory with the same name as the script (without the .py).
# - If --script-arg is "verify" then the responses will be checked against
#   responses saved from a previous "record" invocation.  If they do not
#   match then the test will stop.
# - If --script-arg is "tolerate" then the responses will be checked against
#   saved responses, but a mismatch will not abort the test.
#
# Optional arguments to get/post:
#
# - ignoreResponse=1 
#   The content of any response will not be recorded or verified.
#
# - neuter=pattern
#   neuter=sequence of patterns
#   Applies to "verify" mode only.  Each regular expression pattern will be
#   searched for within the text of the responses.  For matches, the first
#   group (bracketed section) will be replaced by a constant.  This allows you
#   to get rid of portions of the HTML that contain things that change from
#   run to run.
class VerifyTest(CompactTest):
    def __init__(self, name):
        CompactTest.__init__(self, name)
        self._htmlDir = self.getName()
        if not os.path.exists(self._htmlDir):
            os.mkdir(self._htmlDir)
        self._mode = self.getScriptArg()
        if self._mode is None:
            handlerClass = "_DoNothing"
        else:
            if self._mode == "verify":
                handlerClass = "_FussyVerifier"
            elif self._mode == "tolerate":
                handlerClass = "_TolerantVerifier"
            elif self._mode == "record":
                handlerClass = "_Recorder"
            else:
                raise "Unknown script-arg: " + self._mode
        self._handler = eval(handlerClass + "(self)")
        self._pausedHandler = None
        self._transCounts = {}
        self._filePrefix = ''
        
    def mode(self): return self._mode

    def pauseRecording(self):
        """If we are recording the script then this pauses recording.
           The call is harmless if we are not recording."""
        if self._handler.__class__ is _Recorder:
            self._pausedHandler = self._handler
            self._handler = _DoNothing(self)

    def resumeRecording(self):
        """If we were paused then this resumes.  Harmless if not paused."""
        if self._pausedHandler is not None:
            self._handler = self._pausedHandler
            self._pausedHandler = None

    def _htmlSavePath(self):
        "Returns the path to which responses should be saved during recording."
        num = self._transCounts[self._filePrefix]
        return "%s/%s%04d.html" % (self._htmlDir, self._filePrefix, num)

    def setFilePrefix(self, prefix):
        """This prefix will be attached to the names of any files recorded.
           File names will be numbered separately for each prefix."""
        self._filePrefix = prefix
        
    def setTransCount(self, count):
        """Call this to manually set the count of transactions for the
           current file prefix.  Do this after 'if' statements so that
           the transactions after the 'if' get consistent file numbers
           regardless of whether the 'if' was taken."""
        assert count >= self._transCounts.get(self._filePrefix, 0)
        self._transCounts[self._filePrefix] = count
        
    def commonRequest(self, method, url, params, respCode, options):
        CompactTest.commonRequest(self, method, url, params, respCode)
        prefix = self._filePrefix
        if not self._transCounts.has_key(prefix):
            self._transCounts[prefix] = 0
        self._transCounts[prefix] = self._transCounts[prefix] + 1    
        if not options.has_key('ignoreResponse'):
            self._handler.handleResponse(self.getResponse(), options)

    def get(self, url, params, respCode, **options):
        self.commonRequest(HttpTestCase.get, url, params, respCode, options)
        
    def post(self, url, params, respCode, **options):
        self.commonRequest(HttpTestCase.post, url, params, respCode, options)

    def postMultiPart(self, url, params, respCode, **options):
        # I only really want multipart for form data.
        self.commonRequest(HttpTestCase.post, url, params, respCode, options)
        

# Helper Classes
class _Handler:
    """One of these is installed into a SnakeTest depending on what mode it
       is being run in."""
    def __init__(self, testCase):
        self._testCase = testCase
        
class _DoNothing(_Handler):
    def handleResponse(self, response, options):
        pass

class _Recorder(_Handler):
    """When installed as the handler for a SnakeTest class, this records
       the responses from the web server."""
    def handleResponse(self, response, options):
        _saveToFile(self._testCase._htmlSavePath(), response)

class _Verifier(_Handler):
    """When installed as the handler for a SnakeTest class, this verifies
       that each response matches one that was recorded earlier."""
    def handleResponse(self, response, options):
        # Read the response we were expecting from a file.
        fp = open(self._testCase._htmlSavePath(), "rb")
        goodResponse = fp.read()
        fp.close()
        
        # Are there any bits of the response that we should clear to allow
        # them to match?  "Neuter" expressions are regular expressions that
        # use a group to specify a portion of the response that should be
        # made constant.  We usually use this to stop variable IDs screwing
        # up the comparison.
        newResponse = response
        if options.has_key('neuter'):
            neuters = options['neuter']
            if type(neuters) is not types.TupleType:
                neuters = (neuters,)
            for neuterPattern in neuters:
                if type(neuterPattern) is types.StringType:
                    neuterPattern = re.compile(neuterPattern)
                def neuter(s, pat):
                    m = pat.search(s)
                    while m is not None:
                        s = s[0:m.start(1)] + "###" + s[m.end(1):]
                        m = pat.search(s, m.end() - (len(m.group(1)) - 3))
                    return s
                newResponse = neuter(newResponse, neuterPattern)
                goodResponse = neuter(goodResponse, neuterPattern)
        
        # Check that what we got is what we expected.
        if goodResponse != newResponse:
            responsePath = self._testCase._htmlSavePath()
            newResponsePath = responsePath+'.new'
            _saveToFile(newResponsePath, response)
            self.reportMismatch(goodResponse, newResponse,
                                responsePath, newResponsePath)

    def reportMismatch(self, goodResponse, newResponse,
                       goodResponsePath, newResponsePath):
        """Raises an error describing the mismatch between the expected response
           and what we got.  We use a cunning HTMLised diff to present the
           results."""
        msg = "HTML did not match: %s %s\n" % ( goodResponsePath, newResponsePath )
        # Build output that makes it easy to see the difference between
        # the two versions.
        lines = ['']
        interestingLines = [0]
        curLine = 0
        def add(s, interesting, curLine, lines=lines,
                interestingLines=interestingLines):
            firstLine = 1
            for line in re.split(r'\r?\n', s):
                if not firstLine:
                    # Must have found a delimiter.
                    curLine = curLine + 1;
                    lines.append('')
                    interestingLines.append(0)
                firstLine = 0
                lines[curLine] = lines[curLine] + line
                if interesting:
                    interestingLines[curLine] = 1
            return curLine
        a, b = diff.html2list(goodResponse), diff.html2list(newResponse)
        s = difflib.SequenceMatcher(None, a, b)
        for e in s.get_opcodes():
            if e[0] == "replace":
                curLine = add('['+''.join(a[e[1]:e[2]]) + ']-->['+''.join(b[e[3]:e[4]])+"]", 1, curLine)
            elif e[0] == "delete":
                curLine = add('--['+ ''.join(a[e[1]:e[2]]) + "]--", 1, curLine)
            elif e[0] == "insert":
                curLine = add('++['+''.join(b[e[3]:e[4]]) + "]++", 1, curLine)
            elif e[0] == "equal":
                curLine = add(''.join(b[e[3]:e[4]]), 0, curLine)
        # Show interesting lines and the lines nearby.  Put an ellipsis
        # ('...') in the spaces in between.
        lastWasEllipsis = 0
        for i in xrange(0, len(lines)):
            l = len(lines)
            if (interestingLines[i] or i >= 1 and interestingLines[i - 1]
                                    or i >= 2 and interestingLines[i - 2]
                                    or i >= 3 and interestingLines[i - 3]
                                  or l-i >= 2 and interestingLines[i + 1]
                                  or l-i >= 3 and interestingLines[i + 2]
                                  or l-i >= 4 and interestingLines[i + 3]):
                if interestingLines[i]:
                    msg = msg + "* "
                else:
                    msg = msg + "  "
                msg = msg + lines[i] + '\n'
                lastWasEllipsis = 0
            else:
                # This is a line of no interest.  We show an ellipsis 
                # between these sections.
                if not lastWasEllipsis:
                    msg = msg + '  ...' + '\n'
                    lastWasEllipsis = 1
        msg = msg + 'to fix: mv %s %s' % (newResponsePath, goodResponsePath)
        self.raiseMismatch(msg)

class _TolerantVerifier(_Verifier):
    def raiseMismatch(self, msg):
        print msg
        
class _FussyVerifier(_Verifier):
    def raiseMismatch(self, msg):
        raise msg

# Helper functions
_alreadySaved = {}  # Help detects erroneous rerecording of files.
def _saveToFile(path, s):
    fp = open(path, "wb")
    fp.write(s)
    fp.close()

